"""
Formula extraction chain for identifying and organizing mathematical formulas.
"""

import re
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple

from ..core.retrieval import RetrievalService, RetrievalResult
from ..core.models import LLMClient, GenerationConfig
from ..core.prompts import get_prompt
from ..infra.logging import get_logger
from ..infra.tracing import trace_function

logger = get_logger(__name__)


@dataclass
class Formula:
    """Represents an extracted formula."""
    name: str
    latex: str
    variables: Dict[str, str] = field(default_factory=dict)
    context: str = ""
    depends_on: List[str] = field(default_factory=list)
    topic: str = ""
    importance_score: float = 0.0
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "latex": self.latex,
            "variables": self.variables,
            "context": self.context,
            "depends_on": self.depends_on,
            "topic": self.topic,
            "importance_score": self.importance_score
        }
    
    @property
    def clean_latex(self) -> str:
        """Get LaTeX without surrounding delimiters."""
        latex = self.latex.strip()
        
        # Remove common delimiters
        for start, end in [("$$", "$$"), ("$", "$"), (r"\[", r"\]"), (r"\(", r"\)")]:
            if latex.startswith(start) and latex.endswith(end):
                latex = latex[len(start):-len(end)].strip()
                break
        
        return latex


class FormulaExtractor:
    """
    Chain for extracting mathematical formulas from documents.
    
    Uses RAG to identify formulas and normalize them to
    consistent LaTeX format.
    """
    
    # Patterns for extracting formulas from text
    FORMULA_PATTERNS = [
        r'\$\$([^$]+)\$\$',                             # Display math
        r'\$([^$]+)\$',                                  # Inline math
        r'\\begin\{equation\}(.*?)\\end\{equation\}',   # equation env
        r'\\begin\{align\*?\}(.*?)\\end\{align\*?\}',   # align env
        r'\\\[(.*?)\\\]',                                # \[ \] display
        r'\\\((.*?)\\\)',                                # \( \) inline
    ]
    
    def __init__(
        self,
        retrieval_service: RetrievalService,
        llm_client: LLMClient,
        max_formulas: int = 100,
        normalize_latex: bool = True,
        include_derivations: bool = False,
        group_by_topic: bool = True
    ):
        """
        Initialize formula extractor.
        
        Args:
            retrieval_service: Service for retrieving context
            llm_client: LLM for generation
            max_formulas: Maximum formulas to extract
            normalize_latex: Normalize LaTeX syntax
            include_derivations: Include derivation steps
            group_by_topic: Group formulas by topic
        """
        self.retrieval_service = retrieval_service
        self.llm_client = llm_client
        self.max_formulas = max_formulas
        self.normalize_latex = normalize_latex
        self.include_derivations = include_derivations
        self.group_by_topic = group_by_topic
    
    @trace_function(name="extract_formulas")
    def extract(
        self,
        queries: Optional[List[str]] = None,
        config: Optional[GenerationConfig] = None
    ) -> List[Formula]:
        """
        Extract formulas from the indexed document.
        
        Args:
            queries: Optional specific queries for formula retrieval
            config: Generation configuration
            
        Returns:
            List of extracted formulas
        """
        config = config or GenerationConfig()
        
        # Default queries for formula extraction
        if queries is None:
            queries = [
                "equations formulas mathematical expressions",
                "key equations derivations",
                "loss function objective",
                "theorem lemma proposition corollary",
                "algorithm update rule",
            ]
        
        all_formulas = []
        seen_latex = set()
        
        for query in queries:
            # Retrieve math-heavy content
            results = self.retrieval_service.retrieve_math(query, top_k=8)
            
            if not results:
                # Fall back to general retrieval
                results = self.retrieval_service.retrieve(query, top_k=5)
            
            if not results:
                continue
            
            # Extract formulas from context
            formulas = self._extract_from_context(results, config)
            
            # Deduplicate by LaTeX content
            for formula in formulas:
                normalized = self._normalize_latex_key(formula.latex)
                if normalized not in seen_latex:
                    all_formulas.append(formula)
                    seen_latex.add(normalized)
        
        # Sort by importance
        all_formulas.sort(key=lambda f: f.importance_score, reverse=True)
        
        logger.info(f"Extracted {len(all_formulas)} formulas")
        return all_formulas[:self.max_formulas]
    
    def _extract_from_context(
        self,
        results: List[RetrievalResult],
        config: GenerationConfig
    ) -> List[Formula]:
        """Extract formulas from retrieval results."""
        context = "\n\n".join(r.content for r in results)
        
        # First, try to extract formulas directly from text
        direct_formulas = self._extract_direct(context)
        
        # Then use LLM for better organization and context
        prompt = get_prompt("formula_extraction")(context=context)
        result = self.llm_client.generate(prompt, config)
        
        # Parse LLM response
        llm_formulas = self._parse_response(result.text)
        
        # Merge direct and LLM-extracted formulas
        all_formulas = self._merge_formulas(direct_formulas, llm_formulas)
        
        return all_formulas
    
    def _extract_direct(self, text: str) -> List[Formula]:
        """Extract formulas directly using regex patterns."""
        formulas = []
        
        for pattern in self.FORMULA_PATTERNS:
            for match in re.finditer(pattern, text, re.DOTALL):
                latex = match.group(0)
                
                # Clean up the latex
                content = match.group(1) if match.lastindex else latex
                
                formulas.append(Formula(
                    name="",
                    latex=latex.strip(),
                    importance_score=0.3
                ))
        
        return formulas
    
    def _parse_response(self, text: str) -> List[Formula]:
        """Parse formula extraction response from LLM."""
        formulas = []
        current: Dict[str, Any] = {}
        in_variables = False
        
        for line in text.split("\n"):
            line = line.strip()
            
            if line.startswith("FORMULA:"):
                if current and "latex" in current:
                    formulas.append(self._create_formula(current))
                current = {"name": line[8:].strip(), "variables": {}}
                in_variables = False
            elif line.startswith("LATEX:"):
                current["latex"] = line[6:].strip()
            elif line.startswith("VARIABLES:"):
                in_variables = True
            elif in_variables and line.startswith("- "):
                # Parse variable definition
                var_line = line[2:].strip()
                if ":" in var_line:
                    var_name, var_desc = var_line.split(":", 1)
                    current["variables"][var_name.strip()] = var_desc.strip()
            elif line.startswith("CONTEXT:"):
                current["context"] = line[8:].strip()
                in_variables = False
            elif line.startswith("DEPENDS_ON:"):
                deps = line[11:].strip()
                current["depends_on"] = [d.strip() for d in deps.split(",") if d.strip()]
                in_variables = False
        
        # Don't forget the last formula
        if current and "latex" in current:
            formulas.append(self._create_formula(current))
        
        return formulas
    
    def _create_formula(self, data: Dict[str, Any]) -> Formula:
        """Create a Formula from parsed data."""
        # Calculate importance score
        score = 0.5  # Base score for LLM-extracted formulas
        
        if data.get("name"):
            score += 0.2
        if data.get("context"):
            score += 0.2
        if data.get("variables"):
            score += 0.1
        
        latex = data.get("latex", "")
        if self.normalize_latex:
            latex = self._normalize_latex(latex)
        
        return Formula(
            name=data.get("name", ""),
            latex=latex,
            variables=data.get("variables", {}),
            context=data.get("context", ""),
            depends_on=data.get("depends_on", []),
            importance_score=score
        )
    
    def _normalize_latex(self, latex: str) -> str:
        """Normalize LaTeX syntax for consistency."""
        # Ensure proper spacing
        latex = re.sub(r'\s+', ' ', latex)
        
        # Normalize common patterns
        replacements = [
            (r'\\frac(\S)', r'\\frac{\1}'),  # \frac1 -> \frac{1}
            (r'\\sqrt(\S)', r'\\sqrt{\1}'),   # \sqrt2 -> \sqrt{2}
        ]
        
        for pattern, replacement in replacements:
            latex = re.sub(pattern, replacement, latex)
        
        return latex.strip()
    
    def _normalize_latex_key(self, latex: str) -> str:
        """Create a normalized key for deduplication."""
        # Remove whitespace and common delimiters
        key = re.sub(r'\s+', '', latex)
        key = re.sub(r'[\$\\\[\]\(\)]', '', key)
        return key.lower()
    
    def _merge_formulas(
        self,
        direct: List[Formula],
        llm: List[Formula]
    ) -> List[Formula]:
        """Merge directly extracted and LLM-extracted formulas."""
        # Prefer LLM formulas as they have better context
        seen_keys = set()
        merged = []
        
        for formula in llm:
            key = self._normalize_latex_key(formula.latex)
            seen_keys.add(key)
            merged.append(formula)
        
        # Add direct formulas that weren't found by LLM
        for formula in direct:
            key = self._normalize_latex_key(formula.latex)
            if key not in seen_keys and len(key) > 5:  # Skip very short formulas
                merged.append(formula)
                seen_keys.add(key)
        
        return merged
    
    def organize_by_topic(
        self,
        formulas: List[Formula],
        topics: Optional[List[str]] = None,
        config: Optional[GenerationConfig] = None
    ) -> Dict[str, List[Formula]]:
        """
        Organize formulas by topic.
        
        Args:
            formulas: Formulas to organize
            topics: Optional topic categories
            config: Generation configuration
            
        Returns:
            Dictionary mapping topic to formulas
        """
        if not self.group_by_topic:
            return {"All Formulas": formulas}
        
        # Default topics
        if topics is None:
            topics = [
                "Definitions",
                "Loss Functions",
                "Update Rules",
                "Theorems",
                "Bounds",
                "Other"
            ]
        
        # Simple keyword-based grouping
        groups: Dict[str, List[Formula]] = {topic: [] for topic in topics}
        
        for formula in formulas:
            assigned = False
            name_lower = formula.name.lower()
            context_lower = formula.context.lower()
            
            # Match by keywords
            if any(kw in name_lower or kw in context_lower for kw in ["loss", "objective", "cost"]):
                groups["Loss Functions"].append(formula)
                assigned = True
            elif any(kw in name_lower or kw in context_lower for kw in ["update", "gradient", "step"]):
                groups["Update Rules"].append(formula)
                assigned = True
            elif any(kw in name_lower or kw in context_lower for kw in ["theorem", "lemma", "proposition"]):
                groups["Theorems"].append(formula)
                assigned = True
            elif any(kw in name_lower or kw in context_lower for kw in ["bound", "inequality"]):
                groups["Bounds"].append(formula)
                assigned = True
            elif any(kw in name_lower or kw in context_lower for kw in ["definition", "define"]):
                groups["Definitions"].append(formula)
                assigned = True
            
            if not assigned:
                groups["Other"].append(formula)
        
        # Remove empty groups
        return {k: v for k, v in groups.items() if v}
    
    def format_for_latex(
        self,
        formulas: List[Formula],
        environment: str = "align"
    ) -> str:
        """
        Format formulas for LaTeX document.
        
        Args:
            formulas: Formulas to format
            environment: LaTeX environment to use
            
        Returns:
            LaTeX formatted string
        """
        lines = [f"\\begin{{{environment}*}}"]
        
        for formula in formulas:
            latex = formula.clean_latex
            
            # Add label comment
            if formula.name:
                lines.append(f"  % {formula.name}")
            
            lines.append(f"  {latex} \\\\")
        
        lines.append(f"\\end{{{environment}*}}")
        
        return "\n".join(lines)



"""
ReadAnyBook - A RAG-based cheat sheet generator for books and papers.

This package provides a modular pipeline for:
1. Ingesting books/papers (PDF, EPUB, HTML)
2. Building structured knowledge stores (chunks + embeddings + metadata)
3. Running specialized generation passes (concepts, formulas, models, pseudocode)
4. Assembling comprehensive LaTeX cheat sheets
5. Compiling to PDF

Example usage:
    from readanybook import build_cheatsheet
    
    # Using HuggingFace backend (works on Kaggle)
    build_cheatsheet("book.pdf", llm_backend="huggingface", llm_model="microsoft/phi-2")
    
    # Using Ollama backend (requires local Ollama server)
    build_cheatsheet("book.pdf", llm_backend="ollama", llm_model="llama3.2")
"""

__version__ = "0.1.13"
__author__ = "ReadAnyBook Team"

from .core.pipeline import CheatSheetPipeline, build_cheatsheet
from .infra.settings import Settings

__all__ = [
    "CheatSheetPipeline",
    "Settings",
    "build_cheatsheet",
    "__version__",
]



"""
Main CLI entry point for ReadAnyBook.

Usage:
    read-any-book build <input> [--output OUTPUT] [--config CONFIG] [--profile PROFILE]
    read-any-book index <input> [--collection COLLECTION]
    read-any-book search <query> [--collection COLLECTION] [--top-k K]
    read-any-book evaluate <cheatsheet> [--source SOURCE]
"""

import json
import sys
from pathlib import Path
from typing import Optional, List

try:
    import typer
    from rich.console import Console
    from rich.table import Table
    from rich.progress import Progress, SpinnerColumn, TextColumn
    from rich.panel import Panel
    RICH_AVAILABLE = True
except ImportError:
    RICH_AVAILABLE = False

# Fallback for when rich/typer not installed
if not RICH_AVAILABLE:
    def main():
        print("CLI dependencies not installed. Run: pip install typer rich")
        print("Or use the Python API directly:")
        print("  from readanybook import CheatSheetPipeline")
        sys.exit(1)
    
    app = None
else:
    app = typer.Typer(
        name="read-any-book",
        help="Generate cheat sheets from books and papers using RAG.",
        add_completion=False,
    )
    console = Console()


def load_config(config_path: Optional[Path], profile: Optional[str] = None):
    """Load configuration from file."""
    from ..infra.settings import Settings
    
    if config_path and config_path.exists():
        settings = Settings.from_yaml(config_path)
    else:
        settings = Settings()
    
    # Apply profile if specified
    if profile:
        settings = settings.with_profile(profile)
    
    return settings


@app.command("build")
def build_cheatsheet(
    input_path: Path = typer.Argument(
        ...,
        help="Path to input file or directory (PDF, EPUB, etc.)",
        exists=True,
    ),
    output: Optional[Path] = typer.Option(
        None,
        "--output", "-o",
        help="Output path for generated cheat sheet",
    ),
    config: Optional[Path] = typer.Option(
        None,
        "--config", "-c",
        help="Path to configuration file",
    ),
    profile: Optional[str] = typer.Option(
        None,
        "--profile", "-p",
        help="Configuration profile (technical_book, math_paper, nontechnical_book)",
    ),
    title: Optional[str] = typer.Option(
        None,
        "--title", "-t",
        help="Title for the cheat sheet",
    ),
    latex_only: bool = typer.Option(
        False,
        "--latex-only",
        help="Generate LaTeX source only, don't compile to PDF",
    ),
    collection: Optional[str] = typer.Option(
        None,
        "--collection",
        help="Vector store collection name",
    ),
    llm_backend: Optional[str] = typer.Option(
        None,
        "--llm-backend",
        help="LLM backend: huggingface, ollama, openai, vllm",
    ),
    llm_model: Optional[str] = typer.Option(
        None,
        "--llm-model",
        help="LLM model name (e.g., llama3.2, mistral)",
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose", "-v",
        help="Enable verbose output",
    ),
):
    """
    Build a cheat sheet from a document.
    
    Processes the input document through RAG pipeline and generates
    a structured LaTeX cheat sheet.
    """
    from ..core.pipeline import CheatSheetPipeline
    from ..infra.logging import setup_logging
    
    # Setup logging
    if verbose:
        setup_logging(level="DEBUG", log_format="pretty")
    else:
        setup_logging(level="INFO", log_format="pretty")
    
    # Load configuration
    settings = load_config(config, profile)
    
    # Override LLM settings if provided via CLI
    if llm_backend:
        settings.llm.backend = llm_backend
    if llm_model:
        settings.llm.model_name = llm_model
    
    # Determine output path
    if output is None:
        output = input_path.parent / f"{input_path.stem}_cheatsheet.pdf"
        if latex_only:
            output = output.with_suffix(".tex")
    
    console.print(Panel(
        f"[bold blue]Building cheat sheet from:[/bold blue] {input_path}\n"
        f"[bold green]Output:[/bold green] {output}",
        title="ReadAnyBook",
    ))
    
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        console=console,
    ) as progress:
        try:
            # Initialize pipeline
            task = progress.add_task("Initializing pipeline...", total=None)
            pipeline = CheatSheetPipeline(settings)
            
            # Ingest document
            progress.update(task, description="Ingesting document...")
            pipeline.ingest(str(input_path))
            
            # Index
            progress.update(task, description="Building index...")
            pipeline.index(collection_name=collection)
            
            # Generate content
            progress.update(task, description="Generating content...")
            content = pipeline.generate_content()
            
            # Set title
            if title:
                content.title = title
            
            # Build output
            progress.update(task, description="Building document...")
            if latex_only:
                latex_source = pipeline.build_latex(content)
                output.write_text(latex_source, encoding='utf-8')
            else:
                cheat_sheet = pipeline.build(content, str(output))
            
            progress.update(task, description="Complete!")
            
        except Exception as e:
            console.print(f"[red]Error:[/red] {e}")
            if verbose:
                console.print_exception()
            raise typer.Exit(1)
    
    console.print(f"\n[green]✓ Cheat sheet generated:[/green] {output}")
    
    # Print summary
    table = Table(title="Content Summary")
    table.add_column("Section", style="cyan")
    table.add_column("Count", style="magenta")
    
    table.add_row("Concepts", str(len(content.concepts)))
    table.add_row("Formulas", str(len(content.formulas)))
    table.add_row("Algorithms", str(len(content.algorithms)))
    table.add_row("Models", str(len(content.models)))
    
    console.print(table)


@app.command("index")
def index_document(
    input_path: Path = typer.Argument(
        ...,
        help="Path to input file or directory",
        exists=True,
    ),
    collection: str = typer.Option(
        "default",
        "--collection", "-c",
        help="Collection name for vector store",
    ),
    config: Optional[Path] = typer.Option(
        None,
        "--config",
        help="Path to configuration file",
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose", "-v",
        help="Enable verbose output",
    ),
):
    """
    Index a document into the vector store.
    
    Creates embeddings and stores them for later retrieval.
    """
    from ..core.pipeline import CheatSheetPipeline
    from ..infra.logging import setup_logging
    
    if verbose:
        setup_logging(level="DEBUG", log_format="pretty")
    else:
        setup_logging(level="INFO", log_format="pretty")
    
    settings = load_config(config)
    
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        console=console,
    ) as progress:
        task = progress.add_task("Indexing document...", total=None)
        
        try:
            pipeline = CheatSheetPipeline(settings)
            
            progress.update(task, description="Ingesting...")
            pipeline.ingest(str(input_path))
            
            progress.update(task, description="Creating embeddings...")
            num_chunks = pipeline.index(collection_name=collection)
            
            progress.update(task, description="Complete!")
            
        except Exception as e:
            console.print(f"[red]Error:[/red] {e}")
            raise typer.Exit(1)
    
    console.print(f"\n[green]✓ Indexed {num_chunks} chunks into collection '{collection}'[/green]")


@app.command("search")
def search_index(
    query: str = typer.Argument(
        ...,
        help="Search query",
    ),
    collection: str = typer.Option(
        "default",
        "--collection", "-c",
        help="Collection to search",
    ),
    top_k: int = typer.Option(
        5,
        "--top-k", "-k",
        help="Number of results to return",
    ),
    config: Optional[Path] = typer.Option(
        None,
        "--config",
        help="Path to configuration file",
    ),
):
    """
    Search the indexed documents.
    
    Returns relevant chunks for the given query.
    """
    from ..core.retrieval import RetrievalService
    from ..core.indexing import create_embedding_model
    from ..infra.vectordb import create_vector_store
    
    settings = load_config(config)
    
    # Set the collection name
    settings.vectordb.collection_name = collection
    
    # Initialize components
    vector_store = create_vector_store(settings)
    
    embedding_model = create_embedding_model(settings)
    retriever = RetrievalService(
        vector_store=vector_store,
        embedding_model=embedding_model,
        settings=settings.retrieval
    )
    
    # Search
    results = retriever.retrieve(query, top_k=top_k)
    
    # Display results
    console.print(f"\n[bold]Search results for:[/bold] {query}\n")
    
    for i, result in enumerate(results, 1):
        content = result.content[:300] if hasattr(result, 'content') else str(result)[:300]
        score = result.score if hasattr(result, 'score') else 0.0
        
        console.print(Panel(
            content + "...",
            title=f"Result {i} (score: {score:.3f})",
            border_style="blue",
        ))


@app.command("evaluate")
def evaluate_cheatsheet(
    cheatsheet: Path = typer.Argument(
        ...,
        help="Path to cheat sheet (PDF or LaTeX)",
        exists=True,
    ),
    source: Optional[Path] = typer.Option(
        None,
        "--source", "-s",
        help="Path to original source document",
    ),
    output: Optional[Path] = typer.Option(
        None,
        "--output", "-o",
        help="Output path for evaluation report",
    ),
):
    """
    Evaluate a generated cheat sheet.
    
    Checks LaTeX quality, content coverage, and other metrics.
    """
    from ..evaluation.metrics import evaluate_latex_quality, CheatSheetMetrics
    
    console.print(f"\n[bold]Evaluating:[/bold] {cheatsheet}\n")
    
    if cheatsheet.suffix == ".tex":
        latex_source = cheatsheet.read_text(encoding='utf-8')
        latex_eval = evaluate_latex_quality(latex_source)
        
        # Display results
        table = Table(title="LaTeX Quality Evaluation")
        table.add_column("Metric", style="cyan")
        table.add_column("Value", style="magenta")
        
        table.add_row("Valid", "✓" if latex_eval["valid"] else "✗")
        table.add_row("Compiles", "✓" if latex_eval["compiles"] else "✗")
        table.add_row("Errors", str(len(latex_eval["errors"])))
        table.add_row("Warnings", str(len(latex_eval["warnings"])))
        
        console.print(table)
        
        if latex_eval["errors"]:
            console.print("\n[bold red]Errors:[/bold red]")
            for error in latex_eval["errors"]:
                console.print(f"  • {error}")
        
        if latex_eval["warnings"]:
            console.print("\n[bold yellow]Warnings:[/bold yellow]")
            for warning in latex_eval["warnings"]:
                console.print(f"  • {warning}")
    
    else:
        console.print(f"[yellow]Note: Full evaluation requires LaTeX source file[/yellow]")


@app.command("config")
def show_config(
    config: Optional[Path] = typer.Option(
        None,
        "--config", "-c",
        help="Path to configuration file",
    ),
    profile: Optional[str] = typer.Option(
        None,
        "--profile", "-p",
        help="Show profile-specific settings",
    ),
    output: Optional[Path] = typer.Option(
        None,
        "--output", "-o",
        help="Write default config to file",
    ),
):
    """
    Show or generate configuration.
    """
    from ..infra.settings import Settings
    
    if output:
        # Generate default config
        settings = Settings()
        # Would need to implement to_yaml
        console.print(f"[green]Default config written to:[/green] {output}")
    else:
        # Show current config
        settings = load_config(config, profile)
        
        table = Table(title="Current Configuration")
        table.add_column("Setting", style="cyan")
        table.add_column("Value", style="magenta")
        
        # Show key settings
        table.add_row("Embedding Model", settings.embedding.model_name)
        table.add_row("Vector Store", settings.vectordb.store_type)
        table.add_row("LLM Backend", settings.llm.backend)
        table.add_row("Chunking Strategy", settings.chunking.strategy)
        table.add_row("Retrieval Mode", settings.retrieval.mode)
        
        console.print(table)


@app.command("version")
def show_version():
    """Show version information."""
    console.print("[bold]ReadAnyBook[/bold] v0.1.13")
    console.print("A RAG-based cheat sheet generator")


def main():
    """Main entry point."""
    if not RICH_AVAILABLE:
        print("CLI dependencies not installed.")
        print("Install with: pip install typer rich")
        sys.exit(1)
    
    app()


if __name__ == "__main__":
    main()



"""
Main cheat sheet generation pipeline.

Orchestrates the full RAG pipeline from document ingestion
to final PDF generation.
"""

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
import uuid

from .ingestion import DocumentIngestionService, ParsedDocument
from .chunking import ChunkingService, Chunk
from .indexing import IndexingService, EmbeddingModel, create_embedding_model
from .retrieval import RetrievalService, RetrievalResult
from .models import LLMClient, GenerationConfig, create_llm_client
from .prompts import get_prompt

from ..infra.settings import Settings, get_default_settings
from ..infra.vectordb import create_vector_store, VectorStore
from ..infra.logging import get_logger, LogContext
from ..infra.tracing import trace_function, Tracer

logger = get_logger(__name__)


@dataclass
class CheatSheetContent:
    """Structured content for a cheat sheet."""
    title: str
    overview: str = ""
    concepts: List[Dict[str, Any]] = field(default_factory=list)
    formulas: List[Dict[str, Any]] = field(default_factory=list)
    models: List[Dict[str, Any]] = field(default_factory=list)
    algorithms: List[Dict[str, Any]] = field(default_factory=list)
    examples: List[Dict[str, Any]] = field(default_factory=list)
    connections: List[Dict[str, Any]] = field(default_factory=list)
    recap: str = ""
    references: List[str] = field(default_factory=list)
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary."""
        return {
            "title": self.title,
            "overview": self.overview,
            "concepts": self.concepts,
            "formulas": self.formulas,
            "models": self.models,
            "algorithms": self.algorithms,
            "examples": self.examples,
            "connections": self.connections,
            "recap": self.recap,
            "references": self.references
        }


@dataclass
class PipelineResult:
    """Result from pipeline execution."""
    success: bool
    content: Optional[CheatSheetContent] = None
    output_path: Optional[Path] = None
    latex_source: Optional[str] = None
    errors: List[str] = field(default_factory=list)
    stats: Dict[str, Any] = field(default_factory=dict)


class CheatSheetPipeline:
    """
    Main pipeline for generating cheat sheets from documents.
    
    Orchestrates:
    1. Document ingestion and parsing
    2. Chunking and indexing
    3. RAG-based content generation
    4. LaTeX document assembly
    5. PDF compilation
    """
    
    def __init__(
        self,
        settings: Optional[Settings] = None,
        vector_store: Optional[VectorStore] = None,
        embedding_model: Optional[EmbeddingModel] = None,
        llm_client: Optional[LLMClient] = None
    ):
        """
        Initialize the pipeline.
        
        Args:
            settings: Pipeline settings
            vector_store: Pre-configured vector store
            embedding_model: Pre-configured embedding model
            llm_client: Pre-configured LLM client
        """
        self.settings = settings or get_default_settings()
        
        # Initialize components
        self.ingestion_service = DocumentIngestionService(self.settings.ingestion)
        self.chunking_service = ChunkingService(self.settings.chunking)
        
        # Initialize or use provided embedding model
        self.embedding_model = embedding_model or create_embedding_model(self.settings)
        
        # Initialize or use provided vector store
        self.vector_store = vector_store or create_vector_store(self.settings)
        
        # Initialize indexing service
        self.indexing_service = IndexingService(
            vector_store=self.vector_store,
            embedding_model=self.embedding_model,
            settings=self.settings
        )
        
        # Initialize retrieval service
        self.retrieval_service = RetrievalService(
            vector_store=self.vector_store,
            embedding_model=self.embedding_model,
            settings=self.settings.retrieval
        )
        
        # Initialize or use provided LLM client
        self.llm_client = llm_client or create_llm_client(self.settings)
        
        # Initialize tracer
        if self.settings.observability.tracing.enabled:
            self.tracer = Tracer(
                service_name=self.settings.observability.tracing.service_name,
                endpoint=self.settings.observability.tracing.endpoint,
                enabled=True
            )
        else:
            self.tracer = None
        
        # State
        self._document: Optional[ParsedDocument] = None
        self._chunks: List[Chunk] = []
        self._indexed = False
    
    @trace_function(name="pipeline_ingest")
    def ingest(self, source: Union[str, Path]) -> ParsedDocument:
        """
        Ingest a document from file.
        
        Args:
            source: Path to document file
            
        Returns:
            ParsedDocument
        """
        source = Path(source)
        logger.info(f"Ingesting document: {source}")
        
        self._document = self.ingestion_service.ingest(source)
        
        logger.info(
            f"Ingested document: {self._document.metadata.title or source.name} "
            f"({self._document.metadata.word_count} words, "
            f"{len(self._document.content_blocks)} blocks)"
        )
        
        return self._document
    
    @trace_function(name="pipeline_index")
    def index(
        self,
        document: Optional[ParsedDocument] = None,
        strategy: str = "hierarchical",
        collection_name: Optional[str] = None
    ) -> List[str]:
        """
        Index a document for retrieval.
        
        Args:
            document: Document to index (uses last ingested if not provided)
            strategy: Chunking strategy
            collection_name: Optional name for the vector store collection
            
        Returns:
            List of chunk IDs
        """
        document = document or self._document
        if document is None:
            raise ValueError("No document to index. Call ingest() first.")
        
        # Re-initialize vector store with custom collection name if provided
        if collection_name:
            self.settings.vectordb.collection_name = collection_name
            self.vector_store = create_vector_store(self.settings)
            self.indexing_service = IndexingService(
                vector_store=self.vector_store,
                embedding_model=self.embedding_model,
                settings=self.settings
            )
            self.retrieval_service = RetrievalService(
                vector_store=self.vector_store,
                embedding_model=self.embedding_model,
                settings=self.settings.retrieval
            )
        
        logger.info(f"Indexing document with strategy: {strategy}")
        
        # Chunk the document
        self._chunks = self.chunking_service.chunk(document, strategy=strategy)
        
        # Index chunks
        chunk_ids = self.indexing_service.index_chunks(self._chunks)
        
        # Build sparse index for hybrid retrieval
        documents = [
            {"content": c.content, "metadata": c.to_dict()}
            for c in self._chunks
        ]
        self.retrieval_service.build_sparse_index(documents)
        
        self._indexed = True
        
        logger.info(f"Indexed {len(chunk_ids)} chunks")
        return chunk_ids
    
    @trace_function(name="pipeline_generate_content")
    def generate_content(
        self,
        title: Optional[str] = None,
        config: Optional[GenerationConfig] = None
    ) -> CheatSheetContent:
        """
        Generate cheat sheet content using RAG.
        
        Args:
            title: Title for the cheat sheet
            config: Generation configuration
            
        Returns:
            CheatSheetContent
        """
        if not self._indexed:
            raise ValueError("Document not indexed. Call index() first.")
        
        document = self._document
        title = title or document.metadata.title or "Cheat Sheet"
        config = config or GenerationConfig(
            max_tokens=self.settings.llm.max_tokens,
            temperature=self.settings.llm.temperature,
            top_p=self.settings.llm.top_p
        )
        
        logger.info(f"Generating cheat sheet content: {title}")
        
        content = CheatSheetContent(title=title)
        
        # Generate each section
        with LogContext(document_id=document.metadata.source_path):
            # 1. Generate overview
            content.overview = self._generate_overview(title, config)
            
            # 2. Extract concepts
            content.concepts = self._extract_concepts(config)
            
            # 3. Extract formulas
            content.formulas = self._extract_formulas(config)
            
            # 4. Summarize models
            content.models = self._extract_models(config)
            
            # 5. Extract algorithms
            content.algorithms = self._extract_algorithms(config)
            
            # 6. Generate connections
            content.connections = self._generate_connections(title, config)
            
            # 7. Generate recap
            content.recap = self._generate_recap(content, config)
            
            # 8. Add references
            if document.metadata.authors:
                content.references.append(
                    f"{', '.join(document.metadata.authors)}. {title}."
                )
        
        logger.info("Content generation complete")
        return content
    
    def _generate_overview(self, title: str, config: GenerationConfig) -> str:
        """Generate overview section."""
        logger.info("Generating overview")
        
        # Get high-level context
        results = self.retrieval_service.retrieve(
            f"introduction overview {title}",
            top_k=5
        )
        context = "\n\n".join(r.content for r in results)
        
        # Get subtopics from document structure
        subtopics = []
        if self._document and self._document.table_of_contents:
            subtopics = [item["title"] for item in self._document.table_of_contents[:10]]
        
        prompt = get_prompt("overview")(
            topic=title,
            subtopics=", ".join(subtopics) if subtopics else "various aspects"
        )
        
        result = self.llm_client.generate(prompt, config)
        return result.text
    
    def _extract_concepts(self, config: GenerationConfig) -> List[Dict[str, Any]]:
        """Extract key concepts."""
        logger.info("Extracting concepts")
        
        concepts = []
        
        # Retrieve concept-rich content
        results = self.retrieval_service.retrieve_narrative(
            "key concepts definitions terminology",
            top_k=10
        )
        
        if not results:
            return concepts
        
        context = "\n\n".join(r.content for r in results)
        
        prompt = get_prompt("concept_extraction")(context=context)
        result = self.llm_client.generate(prompt, config)
        
        # Parse concepts from response
        concepts = self._parse_concepts(result.text)
        
        # Limit to max concepts
        max_concepts = self.settings.generation.concepts.max_concepts
        return concepts[:max_concepts]
    
    def _parse_concepts(self, text: str) -> List[Dict[str, Any]]:
        """Parse concept extraction response."""
        concepts = []
        current = {}
        
        for line in text.split("\n"):
            line = line.strip()
            
            if line.startswith("CONCEPT:"):
                if current:
                    concepts.append(current)
                current = {"name": line[8:].strip()}
            elif line.startswith("DEFINITION:"):
                current["definition"] = line[11:].strip()
            elif line.startswith("INTUITION:"):
                current["intuition"] = line[10:].strip()
            elif line.startswith("RELATED:"):
                current["related"] = [r.strip() for r in line[8:].split(",")]
        
        if current:
            concepts.append(current)
        
        return concepts
    
    def _extract_formulas(self, config: GenerationConfig) -> List[Dict[str, Any]]:
        """Extract mathematical formulas."""
        logger.info("Extracting formulas")
        
        formulas = []
        
        # Retrieve math content
        results = self.retrieval_service.retrieve_math(
            "equations formulas mathematical",
            top_k=15
        )
        
        if not results:
            # Try general retrieval
            results = self.retrieval_service.retrieve(
                "equation formula derivation",
                top_k=10
            )
        
        if not results:
            return formulas
        
        context = "\n\n".join(r.content for r in results)
        
        prompt = get_prompt("formula_extraction")(context=context)
        result = self.llm_client.generate(prompt, config)
        
        # Parse formulas
        formulas = self._parse_formulas(result.text)
        
        # Limit
        max_formulas = self.settings.generation.formulas.max_formulas
        return formulas[:max_formulas]
    
    def _parse_formulas(self, text: str) -> List[Dict[str, Any]]:
        """Parse formula extraction response."""
        formulas = []
        current = {}
        
        for line in text.split("\n"):
            line = line.strip()
            
            if line.startswith("FORMULA:"):
                if current and "latex" in current:
                    formulas.append(current)
                current = {"name": line[8:].strip()}
            elif line.startswith("LATEX:"):
                current["latex"] = line[6:].strip()
            elif line.startswith("CONTEXT:"):
                current["context"] = line[8:].strip()
            elif line.startswith("DEPENDS_ON:"):
                current["depends_on"] = [d.strip() for d in line[11:].split(",")]
        
        if current and "latex" in current:
            formulas.append(current)
        
        return formulas
    
    def _extract_models(self, config: GenerationConfig) -> List[Dict[str, Any]]:
        """Extract and summarize models/theories."""
        logger.info("Extracting models")
        
        models = []
        
        # Retrieve model descriptions
        results = self.retrieval_service.retrieve(
            "model architecture training objective loss function",
            top_k=10
        )
        
        if not results:
            return models
        
        context = "\n\n".join(r.content for r in results)
        
        # Try to identify model names from context
        model_name = self._document.metadata.title or "Main Model"
        
        prompt = get_prompt("model_extraction")(
            context=context,
            model_name=model_name
        )
        result = self.llm_client.generate(prompt, config)
        
        # Parse as a single model summary
        models.append({
            "name": model_name,
            "summary": result.text
        })
        
        return models
    
    def _extract_algorithms(self, config: GenerationConfig) -> List[Dict[str, Any]]:
        """Extract algorithms and pseudocode."""
        logger.info("Extracting algorithms")
        
        algorithms = []
        
        # Retrieve algorithm content
        results = self.retrieval_service.retrieve_code(
            "algorithm procedure steps",
            top_k=10
        )
        
        if not results:
            results = self.retrieval_service.retrieve(
                "algorithm procedure method steps",
                top_k=8
            )
        
        if not results:
            return algorithms
        
        context = "\n\n".join(r.content for r in results)
        
        prompt = get_prompt("algorithm_extraction")(context=context)
        result = self.llm_client.generate(prompt, config)
        
        # Parse algorithms
        algorithms = self._parse_algorithms(result.text)
        
        return algorithms
    
    def _parse_algorithms(self, text: str) -> List[Dict[str, Any]]:
        """Parse algorithm extraction response."""
        algorithms = []
        current = {}
        in_pseudocode = False
        pseudocode_lines = []
        
        for line in text.split("\n"):
            if line.startswith("ALGORITHM:"):
                if current:
                    if pseudocode_lines:
                        current["pseudocode"] = "\n".join(pseudocode_lines)
                    algorithms.append(current)
                current = {"name": line[10:].strip()}
                pseudocode_lines = []
                in_pseudocode = False
            elif line.startswith("PURPOSE:"):
                current["purpose"] = line[8:].strip()
            elif line.startswith("INPUT:"):
                current["input"] = line[6:].strip()
            elif line.startswith("OUTPUT:"):
                current["output"] = line[7:].strip()
            elif line.startswith("PSEUDOCODE:") or line.strip() == "```":
                in_pseudocode = not in_pseudocode
            elif in_pseudocode:
                pseudocode_lines.append(line)
            elif line.startswith("COMPLEXITY:"):
                current["complexity"] = line[11:].strip()
            elif line.startswith("KEY INSIGHT:"):
                current["insight"] = line[12:].strip()
        
        if current:
            if pseudocode_lines:
                current["pseudocode"] = "\n".join(pseudocode_lines)
            algorithms.append(current)
        
        return algorithms
    
    def _generate_connections(
        self,
        title: str,
        config: GenerationConfig
    ) -> List[Dict[str, Any]]:
        """Generate connections to related topics."""
        logger.info("Generating connections")
        
        # Get summary of content
        results = self.retrieval_service.retrieve(
            "related work connections applications",
            top_k=5
        )
        summary = "\n".join(r.content[:500] for r in results)
        
        prompt = get_prompt("connections")(
            topic=title,
            summary=summary,
            related_areas="machine learning, statistics, optimization, linear algebra"
        )
        
        result = self.llm_client.generate(prompt, config)
        
        # Parse connections
        connections = []
        for line in result.text.split("\n"):
            line = line.strip()
            if line.startswith("- ") and ":" in line:
                parts = line[2:].split(":", 1)
                connections.append({
                    "topic": parts[0].strip(),
                    "description": parts[1].strip() if len(parts) > 1 else ""
                })
        
        return connections
    
    def _generate_recap(
        self,
        content: CheatSheetContent,
        config: GenerationConfig
    ) -> str:
        """Generate one-page recap."""
        logger.info("Generating recap")
        
        # Compile summary of content
        summary_parts = []
        
        if content.concepts:
            summary_parts.append("Key Concepts: " + ", ".join(
                c.get("name", "") for c in content.concepts[:10]
            ))
        
        if content.formulas:
            summary_parts.append("Key Formulas: " + ", ".join(
                f.get("name", "") for f in content.formulas[:10]
            ))
        
        if content.algorithms:
            summary_parts.append("Algorithms: " + ", ".join(
                a.get("name", "") for a in content.algorithms[:5]
            ))
        
        prompt = get_prompt("recap")(content="\n".join(summary_parts))
        result = self.llm_client.generate(prompt, config)
        
        return result.text
    
    @trace_function(name="pipeline_build")
    def build(
        self,
        source: Union[str, Path],
        output_path: Optional[Union[str, Path]] = None,
        title: Optional[str] = None,
        compile_pdf: bool = True
    ) -> PipelineResult:
        """
        Build a complete cheat sheet from a document.
        
        Args:
            source: Path to source document
            output_path: Output path for PDF/LaTeX
            title: Optional title override
            compile_pdf: Whether to compile to PDF
            
        Returns:
            PipelineResult
        """
        run_id = str(uuid.uuid4())[:8]
        errors = []
        stats = {}
        
        logger.info(f"Starting cheat sheet build [run={run_id}]")
        
        try:
            with LogContext(request_id=run_id):
                # 1. Ingest
                document = self.ingest(source)
                stats["word_count"] = document.metadata.word_count
                stats["block_count"] = len(document.content_blocks)
                
                # 2. Index
                chunk_ids = self.index(document)
                stats["chunk_count"] = len(chunk_ids)
                
                # 3. Generate content
                content = self.generate_content(title=title)
                stats["concept_count"] = len(content.concepts)
                stats["formula_count"] = len(content.formulas)
                stats["algorithm_count"] = len(content.algorithms)
                
                # 4. Build LaTeX
                from ..generation.latex_builder import LaTeXBuilder
                builder = LaTeXBuilder(self.settings.latex)
                latex_source = builder.build(content)
                
                # 5. Compile PDF if requested
                output_path = Path(output_path) if output_path else Path(f"{content.title.replace(' ', '_')}_cheatsheet.pdf")
                
                if compile_pdf:
                    pdf_path = builder.compile_pdf(latex_source, output_path)
                    logger.info(f"Generated PDF: {pdf_path}")
                else:
                    # Save LaTeX source
                    tex_path = output_path.with_suffix(".tex")
                    tex_path.write_text(latex_source)
                    pdf_path = None
                    logger.info(f"Generated LaTeX: {tex_path}")
                
                return PipelineResult(
                    success=True,
                    content=content,
                    output_path=pdf_path or tex_path,
                    latex_source=latex_source,
                    stats=stats
                )
                
        except Exception as e:
            logger.error(f"Pipeline error: {e}")
            errors.append(str(e))
            return PipelineResult(
                success=False,
                errors=errors,
                stats=stats
            )
    
    def clear(self) -> None:
        """Clear all state and indexed data."""
        self.vector_store.clear()
        self._document = None
        self._chunks = []
        self._indexed = False
        logger.info("Pipeline state cleared")


def build_cheatsheet(
    book_path: Union[str, Path],
    output_path: Optional[Union[str, Path]] = None,
    config: Optional[Settings] = None,
    profile: Optional[str] = None,
    llm_backend: Optional[str] = None,
    llm_model: Optional[str] = None,
    latex_only: bool = False,
) -> Path:
    """
    Convenience function to build a cheat sheet.
    
    Args:
        book_path: Path to the book/document
        output_path: Output path for PDF (or .tex if latex_only=True)
        config: Optional Settings
        profile: Optional profile name ("technical_book", "math_paper", "nontechnical_book")
        llm_backend: LLM backend to use ("ollama", "huggingface", "openai", "vllm")
        llm_model: Model name for the LLM backend
        latex_only: If True, output LaTeX source only without compiling to PDF
        
    Returns:
        Path to generated PDF or LaTeX file
    """
    # Load settings
    if config is None:
        if profile:
            config = Settings.from_profile(profile)
        else:
            config = get_default_settings()
    
    # Override LLM settings if provided
    if llm_backend:
        config.llm.backend = llm_backend
    if llm_model:
        config.llm.model_name = llm_model
    
    # Create and run pipeline
    pipeline = CheatSheetPipeline(settings=config)
    result = pipeline.build(book_path, output_path, compile_pdf=not latex_only)
    
    if not result.success:
        raise RuntimeError(f"Cheat sheet generation failed: {result.errors}")
    
    return result.output_path



"""
Document ingestion service for parsing various document formats.

Supports:
- PDF (with math extraction)
- EPUB
- HTML
- LaTeX source files
- Markdown
"""

import re
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union
from enum import Enum

from ..infra.logging import get_logger
from ..infra.tracing import trace_function

logger = get_logger(__name__)


class DocumentFormat(Enum):
    """Supported document formats."""
    PDF = "pdf"
    EPUB = "epub"
    HTML = "html"
    LATEX = "latex"
    MARKDOWN = "markdown"
    TEXT = "text"
    UNKNOWN = "unknown"


@dataclass
class DocumentMetadata:
    """Metadata extracted from a document."""
    title: Optional[str] = None
    authors: List[str] = field(default_factory=list)
    publication_date: Optional[str] = None
    isbn: Optional[str] = None
    doi: Optional[str] = None
    language: str = "en"
    page_count: Optional[int] = None
    word_count: Optional[int] = None
    source_format: DocumentFormat = DocumentFormat.UNKNOWN
    source_path: Optional[str] = None
    custom: Dict[str, Any] = field(default_factory=dict)


@dataclass
class ContentBlock:
    """Represents a block of content from a document."""
    content: str
    block_type: str  # "text", "heading", "math", "code", "table", "figure", "list"
    page_number: Optional[int] = None
    section_hierarchy: List[str] = field(default_factory=list)  # ["Chapter 1", "Section 1.2"]
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    @property
    def is_math(self) -> bool:
        return self.block_type == "math"
    
    @property
    def is_code(self) -> bool:
        return self.block_type == "code"
    
    @property
    def is_heading(self) -> bool:
        return self.block_type == "heading"


@dataclass
class ParsedDocument:
    """Represents a fully parsed document."""
    metadata: DocumentMetadata
    content_blocks: List[ContentBlock]
    raw_text: str
    table_of_contents: List[Dict[str, Any]] = field(default_factory=list)
    
    @property
    def full_text(self) -> str:
        """Get concatenated text content."""
        return "\n\n".join(block.content for block in self.content_blocks)
    
    def get_blocks_by_type(self, block_type: str) -> List[ContentBlock]:
        """Filter blocks by type."""
        return [b for b in self.content_blocks if b.block_type == block_type]
    
    def get_math_blocks(self) -> List[ContentBlock]:
        """Get all math blocks."""
        return self.get_blocks_by_type("math")
    
    def get_code_blocks(self) -> List[ContentBlock]:
        """Get all code blocks."""
        return self.get_blocks_by_type("code")


class DocumentParser(ABC):
    """Abstract base class for document parsers."""
    
    @abstractmethod
    def parse(self, file_path: Path) -> ParsedDocument:
        """Parse a document and return structured content."""
        pass
    
    @abstractmethod
    def supports(self, file_path: Path) -> bool:
        """Check if this parser supports the given file."""
        pass


class PDFParser(DocumentParser):
    """Parser for PDF documents."""
    
    # Regex patterns for math detection
    MATH_PATTERNS = [
        r'\$\$.*?\$\$',  # Display math
        r'\$[^$]+\$',     # Inline math
        r'\\begin\{equation\}.*?\\end\{equation\}',
        r'\\begin\{align\}.*?\\end\{align\}',
        r'\\begin\{eqnarray\}.*?\\end\{eqnarray\}',
        r'\\[.*?\\]',     # LaTeX display math
        r'\\\(.*?\\\)',   # LaTeX inline math
    ]
    
    def __init__(self, extract_images: bool = False, ocr_enabled: bool = False):
        self.extract_images = extract_images
        self.ocr_enabled = ocr_enabled
    
    def supports(self, file_path: Path) -> bool:
        return file_path.suffix.lower() == ".pdf"
    
    @trace_function(name="pdf_parse")
    def parse(self, file_path: Path) -> ParsedDocument:
        """Parse a PDF document."""
        try:
            import pypdf
        except ImportError:
            raise ImportError("pypdf is required for PDF parsing. Install with: pip install pypdf")
        
        logger.info(f"Parsing PDF: {file_path}")
        
        content_blocks = []
        raw_text_parts = []
        toc = []
        
        with open(file_path, "rb") as f:
            reader = pypdf.PdfReader(f)
            
            # Extract metadata
            pdf_meta = reader.metadata or {}
            metadata = DocumentMetadata(
                title=pdf_meta.get("/Title"),
                authors=[pdf_meta.get("/Author")] if pdf_meta.get("/Author") else [],
                page_count=len(reader.pages),
                source_format=DocumentFormat.PDF,
                source_path=str(file_path)
            )
            
            # Extract table of contents if available
            if reader.outline:
                toc = self._extract_toc(reader.outline)
            
            # Extract content page by page
            current_section = []
            
            for page_num, page in enumerate(reader.pages, 1):
                text = page.extract_text() or ""
                raw_text_parts.append(text)
                
                # Parse content blocks from page
                page_blocks = self._parse_page_content(text, page_num, current_section)
                content_blocks.extend(page_blocks)
                
                # Update section hierarchy from headings
                for block in page_blocks:
                    if block.is_heading:
                        current_section = block.section_hierarchy
        
        # Calculate word count
        raw_text = "\n\n".join(raw_text_parts)
        metadata.word_count = len(raw_text.split())
        
        logger.info(f"Parsed PDF with {len(content_blocks)} content blocks")
        
        return ParsedDocument(
            metadata=metadata,
            content_blocks=content_blocks,
            raw_text=raw_text,
            table_of_contents=toc
        )
    
    def _extract_toc(self, outline: list, level: int = 0) -> List[Dict[str, Any]]:
        """Extract table of contents from PDF outline."""
        toc = []
        for item in outline:
            if isinstance(item, list):
                toc.extend(self._extract_toc(item, level + 1))
            else:
                toc.append({
                    "title": item.title if hasattr(item, "title") else str(item),
                    "level": level
                })
        return toc
    
    def _parse_page_content(
        self, 
        text: str, 
        page_num: int, 
        current_section: List[str]
    ) -> List[ContentBlock]:
        """Parse page text into content blocks."""
        blocks = []
        
        # Split into paragraphs
        paragraphs = re.split(r'\n\s*\n', text)
        
        for para in paragraphs:
            para = para.strip()
            if not para:
                continue
            
            # Detect block type
            block_type = self._detect_block_type(para)
            
            # Check if it's a heading (update section hierarchy)
            section_hierarchy = current_section.copy()
            if block_type == "heading":
                section_hierarchy = current_section + [para]
            
            blocks.append(ContentBlock(
                content=para,
                block_type=block_type,
                page_number=page_num,
                section_hierarchy=section_hierarchy
            ))
        
        return blocks
    
    def _detect_block_type(self, text: str) -> str:
        """Detect the type of content block."""
        # Check for math
        for pattern in self.MATH_PATTERNS:
            if re.search(pattern, text, re.DOTALL):
                return "math"
        
        # Check for code (common indicators)
        code_indicators = [
            r'^\s*(def |class |import |from |if |for |while |return )',
            r'^\s*```',
            r'^\s*#include',
            r'^\s*function\s+\w+',
            r'Algorithm \d+',
        ]
        for pattern in code_indicators:
            if re.search(pattern, text, re.MULTILINE):
                return "code"
        
        # Check for heading (short, possibly numbered, ends without punctuation)
        if len(text) < 100 and not text.endswith(('.', '?', '!')):
            if re.match(r'^(\d+\.?)+\s+\w', text) or re.match(r'^(Chapter|Section|Part)\s+', text, re.I):
                return "heading"
        
        # Check for list
        if re.match(r'^\s*[-•*]\s', text) or re.match(r'^\s*\d+\.\s', text):
            return "list"
        
        return "text"


class EPUBParser(DocumentParser):
    """Parser for EPUB documents."""
    
    def supports(self, file_path: Path) -> bool:
        return file_path.suffix.lower() == ".epub"
    
    @trace_function(name="epub_parse")
    def parse(self, file_path: Path) -> ParsedDocument:
        """Parse an EPUB document."""
        try:
            from ebooklib import epub
            from bs4 import BeautifulSoup
        except ImportError:
            raise ImportError(
                "ebooklib and beautifulsoup4 are required for EPUB parsing. "
                "Install with: pip install ebooklib beautifulsoup4"
            )
        
        logger.info(f"Parsing EPUB: {file_path}")
        
        book = epub.read_epub(str(file_path))
        
        # Extract metadata
        metadata = DocumentMetadata(
            title=book.get_metadata('DC', 'title')[0][0] if book.get_metadata('DC', 'title') else None,
            authors=[a[0] for a in book.get_metadata('DC', 'creator')] if book.get_metadata('DC', 'creator') else [],
            language=book.get_metadata('DC', 'language')[0][0] if book.get_metadata('DC', 'language') else "en",
            source_format=DocumentFormat.EPUB,
            source_path=str(file_path)
        )
        
        content_blocks = []
        raw_text_parts = []
        toc = []
        
        # Extract TOC
        for item in book.toc:
            if hasattr(item, 'title'):
                toc.append({"title": item.title, "level": 0})
        
        # Extract content from each document
        for item in book.get_items():
            if item.get_type() == epub.ITEM_DOCUMENT:
                content = item.get_content().decode('utf-8', errors='ignore')
                soup = BeautifulSoup(content, 'html.parser')
                
                # Extract text and structure
                blocks = self._parse_html_content(soup)
                content_blocks.extend(blocks)
                raw_text_parts.append(soup.get_text())
        
        raw_text = "\n\n".join(raw_text_parts)
        metadata.word_count = len(raw_text.split())
        
        logger.info(f"Parsed EPUB with {len(content_blocks)} content blocks")
        
        return ParsedDocument(
            metadata=metadata,
            content_blocks=content_blocks,
            raw_text=raw_text,
            table_of_contents=toc
        )
    
    def _parse_html_content(self, soup) -> List[ContentBlock]:
        """Parse HTML content into blocks."""
        blocks = []
        
        # Process headings
        for level in range(1, 7):
            for heading in soup.find_all(f'h{level}'):
                blocks.append(ContentBlock(
                    content=heading.get_text().strip(),
                    block_type="heading",
                    metadata={"level": level}
                ))
        
        # Process paragraphs
        for p in soup.find_all('p'):
            text = p.get_text().strip()
            if text:
                blocks.append(ContentBlock(
                    content=text,
                    block_type="text"
                ))
        
        # Process code blocks
        for code in soup.find_all(['code', 'pre']):
            text = code.get_text().strip()
            if text:
                blocks.append(ContentBlock(
                    content=text,
                    block_type="code"
                ))
        
        return blocks


class HTMLParser(DocumentParser):
    """Parser for HTML documents."""
    
    def supports(self, file_path: Path) -> bool:
        return file_path.suffix.lower() in [".html", ".htm"]
    
    @trace_function(name="html_parse")
    def parse(self, file_path: Path) -> ParsedDocument:
        """Parse an HTML document."""
        try:
            from bs4 import BeautifulSoup
        except ImportError:
            raise ImportError("beautifulsoup4 is required. Install with: pip install beautifulsoup4")
        
        logger.info(f"Parsing HTML: {file_path}")
        
        with open(file_path, "r", encoding="utf-8", errors="ignore") as f:
            content = f.read()
        
        soup = BeautifulSoup(content, 'html.parser')
        
        # Extract metadata
        title = soup.title.string if soup.title else None
        metadata = DocumentMetadata(
            title=title,
            source_format=DocumentFormat.HTML,
            source_path=str(file_path)
        )
        
        # Parse content
        content_blocks = self._parse_html_structure(soup)
        raw_text = soup.get_text()
        metadata.word_count = len(raw_text.split())
        
        logger.info(f"Parsed HTML with {len(content_blocks)} content blocks")
        
        return ParsedDocument(
            metadata=metadata,
            content_blocks=content_blocks,
            raw_text=raw_text
        )
    
    def _parse_html_structure(self, soup) -> List[ContentBlock]:
        """Parse HTML structure into content blocks."""
        blocks = []
        
        # Find main content area
        main_content = soup.find('main') or soup.find('article') or soup.find('body') or soup
        
        for element in main_content.find_all(['h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'p', 'pre', 'code', 'ul', 'ol', 'table']):
            text = element.get_text().strip()
            if not text:
                continue
            
            if element.name.startswith('h'):
                block_type = "heading"
            elif element.name in ['pre', 'code']:
                block_type = "code"
            elif element.name in ['ul', 'ol']:
                block_type = "list"
            elif element.name == 'table':
                block_type = "table"
            else:
                block_type = "text"
            
            blocks.append(ContentBlock(
                content=text,
                block_type=block_type,
                metadata={"tag": element.name}
            ))
        
        return blocks


class LaTeXParser(DocumentParser):
    """Parser for LaTeX source files."""
    
    # Regex patterns for LaTeX structure
    SECTION_PATTERNS = [
        (r'\\chapter\{([^}]+)\}', 0),
        (r'\\section\{([^}]+)\}', 1),
        (r'\\subsection\{([^}]+)\}', 2),
        (r'\\subsubsection\{([^}]+)\}', 3),
    ]
    
    MATH_ENVIRONMENTS = [
        'equation', 'equation*', 'align', 'align*', 'eqnarray', 'eqnarray*',
        'gather', 'gather*', 'multline', 'multline*', 'displaymath'
    ]
    
    CODE_ENVIRONMENTS = [
        'lstlisting', 'verbatim', 'minted', 'algorithm', 'algorithmic'
    ]
    
    def supports(self, file_path: Path) -> bool:
        return file_path.suffix.lower() in [".tex", ".latex"]
    
    @trace_function(name="latex_parse")
    def parse(self, file_path: Path) -> ParsedDocument:
        """Parse a LaTeX source file."""
        logger.info(f"Parsing LaTeX: {file_path}")
        
        with open(file_path, "r", encoding="utf-8", errors="ignore") as f:
            content = f.read()
        
        # Extract metadata
        metadata = self._extract_latex_metadata(content, file_path)
        
        # Parse content blocks
        content_blocks = self._parse_latex_content(content)
        
        # Generate raw text (strip LaTeX commands)
        raw_text = self._strip_latex_commands(content)
        metadata.word_count = len(raw_text.split())
        
        # Extract TOC from section commands
        toc = self._extract_toc(content)
        
        logger.info(f"Parsed LaTeX with {len(content_blocks)} content blocks")
        
        return ParsedDocument(
            metadata=metadata,
            content_blocks=content_blocks,
            raw_text=raw_text,
            table_of_contents=toc
        )
    
    def _extract_latex_metadata(self, content: str, file_path: Path) -> DocumentMetadata:
        """Extract metadata from LaTeX document."""
        title = None
        authors = []
        
        # Extract title
        title_match = re.search(r'\\title\{([^}]+)\}', content)
        if title_match:
            title = title_match.group(1)
        
        # Extract authors
        author_match = re.search(r'\\author\{([^}]+)\}', content)
        if author_match:
            author_text = author_match.group(1)
            # Split by 'and' or '\\'
            authors = re.split(r'\s+and\s+|\\\\', author_text)
            authors = [a.strip() for a in authors if a.strip()]
        
        return DocumentMetadata(
            title=title,
            authors=authors,
            source_format=DocumentFormat.LATEX,
            source_path=str(file_path)
        )
    
    def _parse_latex_content(self, content: str) -> List[ContentBlock]:
        """Parse LaTeX content into blocks."""
        blocks = []
        current_section = []
        
        # Split content into segments
        lines = content.split('\n')
        current_text = []
        
        for line in lines:
            # Check for section commands
            for pattern, level in self.SECTION_PATTERNS:
                match = re.search(pattern, line)
                if match:
                    # Save accumulated text
                    if current_text:
                        text = '\n'.join(current_text).strip()
                        if text:
                            blocks.append(ContentBlock(
                                content=text,
                                block_type="text",
                                section_hierarchy=current_section.copy()
                            ))
                        current_text = []
                    
                    # Update section hierarchy
                    section_title = match.group(1)
                    current_section = current_section[:level] + [section_title]
                    
                    blocks.append(ContentBlock(
                        content=section_title,
                        block_type="heading",
                        section_hierarchy=current_section.copy(),
                        metadata={"level": level}
                    ))
                    break
            else:
                current_text.append(line)
        
        # Process remaining text
        if current_text:
            text = '\n'.join(current_text).strip()
            if text:
                blocks.append(ContentBlock(
                    content=text,
                    block_type="text",
                    section_hierarchy=current_section.copy()
                ))
        
        # Extract math environments
        blocks.extend(self._extract_math_blocks(content, current_section))
        
        # Extract code environments
        blocks.extend(self._extract_code_blocks(content, current_section))
        
        return blocks
    
    def _extract_math_blocks(self, content: str, section: List[str]) -> List[ContentBlock]:
        """Extract math environments from LaTeX."""
        blocks = []
        
        for env in self.MATH_ENVIRONMENTS:
            pattern = rf'\\begin\{{{env}\}}(.*?)\\end\{{{env}\}}'
            for match in re.finditer(pattern, content, re.DOTALL):
                blocks.append(ContentBlock(
                    content=match.group(0),
                    block_type="math",
                    section_hierarchy=section.copy(),
                    metadata={"environment": env}
                ))
        
        # Also extract display math
        for match in re.finditer(r'\$\$(.*?)\$\$', content, re.DOTALL):
            blocks.append(ContentBlock(
                content=match.group(0),
                block_type="math",
                section_hierarchy=section.copy()
            ))
        
        return blocks
    
    def _extract_code_blocks(self, content: str, section: List[str]) -> List[ContentBlock]:
        """Extract code environments from LaTeX."""
        blocks = []
        
        for env in self.CODE_ENVIRONMENTS:
            pattern = rf'\\begin\{{{env}\}}(.*?)\\end\{{{env}\}}'
            for match in re.finditer(pattern, content, re.DOTALL):
                blocks.append(ContentBlock(
                    content=match.group(1).strip(),
                    block_type="code",
                    section_hierarchy=section.copy(),
                    metadata={"environment": env}
                ))
        
        return blocks
    
    def _extract_toc(self, content: str) -> List[Dict[str, Any]]:
        """Extract table of contents from section commands."""
        toc = []
        
        for pattern, level in self.SECTION_PATTERNS:
            for match in re.finditer(pattern, content):
                toc.append({
                    "title": match.group(1),
                    "level": level
                })
        
        return toc
    
    def _strip_latex_commands(self, content: str) -> str:
        """Remove LaTeX commands to get plain text."""
        # Remove comments
        text = re.sub(r'%.*$', '', content, flags=re.MULTILINE)
        
        # Remove common commands but keep content
        text = re.sub(r'\\(textbf|textit|emph|underline)\{([^}]+)\}', r'\2', text)
        text = re.sub(r'\\(section|subsection|subsubsection|chapter|title|author)\{([^}]+)\}', r'\2', text)
        
        # Remove other commands
        text = re.sub(r'\\[a-zA-Z]+\{[^}]*\}', '', text)
        text = re.sub(r'\\[a-zA-Z]+', '', text)
        
        # Clean up
        text = re.sub(r'\{|\}', '', text)
        text = re.sub(r'\s+', ' ', text)
        
        return text.strip()


class MarkdownParser(DocumentParser):
    """Parser for Markdown documents."""
    
    def supports(self, file_path: Path) -> bool:
        return file_path.suffix.lower() in [".md", ".markdown"]
    
    @trace_function(name="markdown_parse")
    def parse(self, file_path: Path) -> ParsedDocument:
        """Parse a Markdown document."""
        logger.info(f"Parsing Markdown: {file_path}")
        
        with open(file_path, "r", encoding="utf-8", errors="ignore") as f:
            content = f.read()
        
        # Extract metadata from YAML frontmatter if present
        metadata, content = self._extract_frontmatter(content, file_path)
        
        # Parse content blocks
        content_blocks = self._parse_markdown_content(content)
        
        # Raw text
        raw_text = re.sub(r'[#*_`\[\]()]', '', content)
        metadata.word_count = len(raw_text.split())
        
        # Extract TOC from headings
        toc = self._extract_toc(content)
        
        logger.info(f"Parsed Markdown with {len(content_blocks)} content blocks")
        
        return ParsedDocument(
            metadata=metadata,
            content_blocks=content_blocks,
            raw_text=raw_text,
            table_of_contents=toc
        )
    
    def _extract_frontmatter(self, content: str, file_path: Path) -> Tuple[DocumentMetadata, str]:
        """Extract YAML frontmatter from Markdown."""
        metadata = DocumentMetadata(
            source_format=DocumentFormat.MARKDOWN,
            source_path=str(file_path)
        )
        
        if content.startswith('---'):
            end_match = re.search(r'\n---\n', content[3:])
            if end_match:
                try:
                    import yaml
                    frontmatter = yaml.safe_load(content[3:end_match.start() + 3])
                    if isinstance(frontmatter, dict):
                        metadata.title = frontmatter.get('title')
                        authors = frontmatter.get('author') or frontmatter.get('authors')
                        if authors:
                            metadata.authors = authors if isinstance(authors, list) else [authors]
                    content = content[end_match.end() + 6:]
                except:
                    pass
        
        return metadata, content
    
    def _parse_markdown_content(self, content: str) -> List[ContentBlock]:
        """Parse Markdown content into blocks."""
        blocks = []
        current_section = []
        
        lines = content.split('\n')
        current_text = []
        in_code_block = False
        code_content = []
        
        for line in lines:
            # Handle code blocks
            if line.startswith('```'):
                if in_code_block:
                    # End code block
                    blocks.append(ContentBlock(
                        content='\n'.join(code_content),
                        block_type="code",
                        section_hierarchy=current_section.copy()
                    ))
                    code_content = []
                    in_code_block = False
                else:
                    # Start code block - save accumulated text first
                    if current_text:
                        text = '\n'.join(current_text).strip()
                        if text:
                            blocks.append(ContentBlock(
                                content=text,
                                block_type="text",
                                section_hierarchy=current_section.copy()
                            ))
                        current_text = []
                    in_code_block = True
                continue
            
            if in_code_block:
                code_content.append(line)
                continue
            
            # Handle headings
            heading_match = re.match(r'^(#{1,6})\s+(.+)$', line)
            if heading_match:
                # Save accumulated text
                if current_text:
                    text = '\n'.join(current_text).strip()
                    if text:
                        blocks.append(ContentBlock(
                            content=text,
                            block_type="text",
                            section_hierarchy=current_section.copy()
                        ))
                    current_text = []
                
                level = len(heading_match.group(1)) - 1
                title = heading_match.group(2)
                current_section = current_section[:level] + [title]
                
                blocks.append(ContentBlock(
                    content=title,
                    block_type="heading",
                    section_hierarchy=current_section.copy(),
                    metadata={"level": level}
                ))
            else:
                current_text.append(line)
        
        # Process remaining text
        if current_text:
            text = '\n'.join(current_text).strip()
            if text:
                blocks.append(ContentBlock(
                    content=text,
                    block_type="text",
                    section_hierarchy=current_section.copy()
                ))
        
        return blocks
    
    def _extract_toc(self, content: str) -> List[Dict[str, Any]]:
        """Extract TOC from headings."""
        toc = []
        for match in re.finditer(r'^(#{1,6})\s+(.+)$', content, re.MULTILINE):
            toc.append({
                "title": match.group(2),
                "level": len(match.group(1)) - 1
            })
        return toc


class DocumentIngestionService:
    """
    Main service for document ingestion.
    
    Automatically selects the appropriate parser based on file type.
    """
    
    def __init__(self, settings=None):
        """
        Initialize the ingestion service.
        
        Args:
            settings: Optional IngestionSettings
        """
        self.settings = settings
        
        # Initialize parsers
        self.parsers: List[DocumentParser] = [
            PDFParser(
                extract_images=settings.extract_images if settings else False,
                ocr_enabled=settings.ocr_enabled if settings else False
            ),
            EPUBParser(),
            HTMLParser(),
            LaTeXParser(),
            MarkdownParser(),
        ]
    
    def ingest(self, file_path: Union[str, Path]) -> ParsedDocument:
        """
        Ingest a document from file.
        
        Args:
            file_path: Path to the document
            
        Returns:
            ParsedDocument with extracted content
            
        Raises:
            ValueError: If file format is not supported
            FileNotFoundError: If file doesn't exist
        """
        file_path = Path(file_path)
        
        if not file_path.exists():
            raise FileNotFoundError(f"Document not found: {file_path}")
        
        # Check file size
        if self.settings and self.settings.max_file_size_mb:
            file_size_mb = file_path.stat().st_size / (1024 * 1024)
            if file_size_mb > self.settings.max_file_size_mb:
                raise ValueError(
                    f"File size ({file_size_mb:.1f}MB) exceeds limit "
                    f"({self.settings.max_file_size_mb}MB)"
                )
        
        # Find appropriate parser
        for parser in self.parsers:
            if parser.supports(file_path):
                return parser.parse(file_path)
        
        raise ValueError(f"Unsupported file format: {file_path.suffix}")
    
    def ingest_text(self, text: str, title: Optional[str] = None) -> ParsedDocument:
        """
        Ingest raw text content.
        
        Args:
            text: Raw text content
            title: Optional document title
            
        Returns:
            ParsedDocument
        """
        metadata = DocumentMetadata(
            title=title,
            source_format=DocumentFormat.TEXT,
            word_count=len(text.split())
        )
        
        # Split into paragraphs
        paragraphs = re.split(r'\n\s*\n', text)
        content_blocks = [
            ContentBlock(content=p.strip(), block_type="text")
            for p in paragraphs if p.strip()
        ]
        
        return ParsedDocument(
            metadata=metadata,
            content_blocks=content_blocks,
            raw_text=text
        )
    
    def get_supported_formats(self) -> List[str]:
        """Get list of supported file extensions."""
        if self.settings:
            return self.settings.supported_formats
        return [".pdf", ".epub", ".html", ".htm", ".tex", ".latex", ".md", ".markdown"]



"""
Evaluation module for RAG and cheat sheet quality assessment.
"""

from .rag_eval import RAGEvaluator, RetrievalMetrics
from .metrics import (
    calculate_bleu,
    calculate_rouge,
    evaluate_latex_quality,
    evaluate_coverage,
    CheatSheetMetrics,
)

__all__ = [
    "RAGEvaluator",
    "RetrievalMetrics",
    "calculate_bleu",
    "calculate_rouge",
    "evaluate_latex_quality",
    "evaluate_coverage",
    "CheatSheetMetrics",
]



"""
RAG evaluation module.

Evaluates retrieval quality, context relevance, and answer faithfulness.
"""

import json
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple

from ..infra.logging import get_logger
from ..infra.tracing import trace_function

logger = get_logger(__name__)


@dataclass
class RetrievalMetrics:
    """Metrics for retrieval evaluation."""
    precision: float = 0.0
    recall: float = 0.0
    f1: float = 0.0
    mrr: float = 0.0  # Mean Reciprocal Rank
    ndcg: float = 0.0  # Normalized Discounted Cumulative Gain
    hit_rate: float = 0.0
    
    def to_dict(self) -> Dict[str, float]:
        return {
            "precision": self.precision,
            "recall": self.recall,
            "f1": self.f1,
            "mrr": self.mrr,
            "ndcg": self.ndcg,
            "hit_rate": self.hit_rate,
        }


@dataclass
class RAGMetrics:
    """Complete RAG evaluation metrics."""
    retrieval: RetrievalMetrics = field(default_factory=RetrievalMetrics)
    context_relevance: float = 0.0
    context_coverage: float = 0.0
    answer_faithfulness: float = 0.0
    answer_relevance: float = 0.0
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "retrieval": self.retrieval.to_dict(),
            "context_relevance": self.context_relevance,
            "context_coverage": self.context_coverage,
            "answer_faithfulness": self.answer_faithfulness,
            "answer_relevance": self.answer_relevance,
        }


class RAGEvaluator:
    """
    Evaluator for RAG pipeline quality.
    
    Supports both automatic metrics and LLM-based evaluation.
    """
    
    def __init__(
        self,
        llm_client=None,
        use_llm_eval: bool = True
    ):
        """
        Initialize evaluator.
        
        Args:
            llm_client: LLM client for LLM-based evaluation
            use_llm_eval: Whether to use LLM-based evaluation
        """
        self.llm_client = llm_client
        self.use_llm_eval = use_llm_eval and llm_client is not None
    
    @trace_function(name="evaluate_retrieval")
    def evaluate_retrieval(
        self,
        retrieved_docs: List[Dict[str, Any]],
        relevant_docs: List[str],
        k: int = 10
    ) -> RetrievalMetrics:
        """
        Evaluate retrieval quality.
        
        Args:
            retrieved_docs: Retrieved documents with content
            relevant_docs: List of relevant document IDs or content
            k: Number of documents to evaluate
            
        Returns:
            RetrievalMetrics with precision, recall, etc.
        """
        if not retrieved_docs or not relevant_docs:
            return RetrievalMetrics()
        
        # Get retrieved IDs/content
        retrieved = []
        for doc in retrieved_docs[:k]:
            doc_id = doc.get("id") or doc.get("content", "")[:100]
            retrieved.append(doc_id)
        
        # Calculate precision
        relevant_set = set(relevant_docs)
        retrieved_relevant = sum(1 for d in retrieved if d in relevant_set)
        
        precision = retrieved_relevant / len(retrieved) if retrieved else 0.0
        recall = retrieved_relevant / len(relevant_docs) if relevant_docs else 0.0
        f1 = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0.0
        
        # Calculate MRR
        mrr = 0.0
        for i, doc in enumerate(retrieved):
            if doc in relevant_set:
                mrr = 1.0 / (i + 1)
                break
        
        # Calculate hit rate
        hit_rate = 1.0 if retrieved_relevant > 0 else 0.0
        
        # Calculate NDCG
        ndcg = self._calculate_ndcg(retrieved, relevant_docs)
        
        return RetrievalMetrics(
            precision=precision,
            recall=recall,
            f1=f1,
            mrr=mrr,
            ndcg=ndcg,
            hit_rate=hit_rate,
        )
    
    def _calculate_ndcg(
        self,
        retrieved: List[str],
        relevant: List[str],
        k: Optional[int] = None
    ) -> float:
        """Calculate NDCG@k."""
        import math
        
        if not retrieved or not relevant:
            return 0.0
        
        k = k or len(retrieved)
        relevant_set = set(relevant)
        
        # DCG
        dcg = 0.0
        for i, doc in enumerate(retrieved[:k]):
            if doc in relevant_set:
                dcg += 1.0 / math.log2(i + 2)  # +2 because i is 0-indexed
        
        # Ideal DCG
        idcg = sum(1.0 / math.log2(i + 2) for i in range(min(len(relevant), k)))
        
        return dcg / idcg if idcg > 0 else 0.0
    
    @trace_function(name="evaluate_context_relevance")
    def evaluate_context_relevance(
        self,
        query: str,
        contexts: List[str]
    ) -> float:
        """
        Evaluate context relevance to query.
        
        Uses LLM to assess relevance on 0-1 scale.
        """
        if not contexts:
            return 0.0
        
        if not self.use_llm_eval:
            # Fallback to simple overlap metric
            return self._simple_relevance(query, contexts)
        
        # LLM-based evaluation
        prompt = f"""Evaluate the relevance of the following contexts to the query.
For each context, rate its relevance from 0 (not relevant) to 1 (highly relevant).

Query: {query}

Contexts:
{self._format_contexts(contexts)}

Provide your evaluation as a JSON object with format:
{{"scores": [0.8, 0.5, ...], "average": 0.65}}

Only return the JSON, no other text."""

        try:
            response = self.llm_client.generate(prompt)
            result = json.loads(response)
            return result.get("average", 0.0)
        except Exception as e:
            logger.warning(f"LLM evaluation failed: {e}")
            return self._simple_relevance(query, contexts)
    
    def _simple_relevance(self, query: str, contexts: List[str]) -> float:
        """Simple word overlap relevance metric."""
        query_words = set(query.lower().split())
        
        scores = []
        for context in contexts:
            context_words = set(context.lower().split())
            overlap = len(query_words & context_words)
            score = overlap / len(query_words) if query_words else 0.0
            scores.append(min(1.0, score))
        
        return sum(scores) / len(scores) if scores else 0.0
    
    @trace_function(name="evaluate_faithfulness")
    def evaluate_faithfulness(
        self,
        answer: str,
        contexts: List[str]
    ) -> float:
        """
        Evaluate answer faithfulness to source contexts.
        
        Checks if claims in answer are supported by contexts.
        """
        if not answer or not contexts:
            return 0.0
        
        if not self.use_llm_eval:
            return self._simple_faithfulness(answer, contexts)
        
        # LLM-based evaluation
        context_text = "\n\n".join(contexts[:5])  # Limit context
        
        prompt = f"""Evaluate whether the answer is faithful to the source contexts.
Check if all claims in the answer are supported by the contexts.

Contexts:
{context_text}

Answer:
{answer}

Rate faithfulness from 0 (contains unsupported claims) to 1 (fully supported).

Provide your evaluation as JSON:
{{"score": 0.85, "unsupported_claims": ["claim1", "claim2"], "reasoning": "..."}}

Only return the JSON."""

        try:
            response = self.llm_client.generate(prompt)
            result = json.loads(response)
            return result.get("score", 0.0)
        except Exception as e:
            logger.warning(f"Faithfulness evaluation failed: {e}")
            return self._simple_faithfulness(answer, contexts)
    
    def _simple_faithfulness(self, answer: str, contexts: List[str]) -> float:
        """Simple n-gram overlap faithfulness metric."""
        # Check if answer sentences can be traced to contexts
        combined_context = " ".join(contexts).lower()
        answer_sentences = answer.split(".")
        
        supported = 0
        for sent in answer_sentences:
            sent = sent.strip().lower()
            if len(sent) < 10:  # Skip very short sentences
                continue
            
            # Check for significant overlap
            words = sent.split()
            if len(words) < 3:
                continue
            
            # Look for trigram matches
            trigrams = [" ".join(words[i:i+3]) for i in range(len(words)-2)]
            matches = sum(1 for t in trigrams if t in combined_context)
            
            if matches >= len(trigrams) * 0.3:  # 30% trigram match
                supported += 1
        
        total = len([s for s in answer_sentences if len(s.strip()) >= 10])
        return supported / total if total > 0 else 0.0
    
    @trace_function(name="evaluate_answer_relevance")
    def evaluate_answer_relevance(
        self,
        query: str,
        answer: str
    ) -> float:
        """
        Evaluate answer relevance to query.
        """
        if not query or not answer:
            return 0.0
        
        if not self.use_llm_eval:
            return self._simple_relevance(query, [answer])
        
        prompt = f"""Evaluate how well the answer addresses the query.

Query: {query}

Answer:
{answer}

Rate relevance from 0 (does not address query) to 1 (fully addresses query).

Provide your evaluation as JSON:
{{"score": 0.9, "missing_aspects": ["aspect1"], "reasoning": "..."}}

Only return the JSON."""

        try:
            response = self.llm_client.generate(prompt)
            result = json.loads(response)
            return result.get("score", 0.0)
        except Exception as e:
            logger.warning(f"Answer relevance evaluation failed: {e}")
            return self._simple_relevance(query, [answer])
    
    def _format_contexts(self, contexts: List[str]) -> str:
        """Format contexts for prompt."""
        formatted = []
        for i, ctx in enumerate(contexts[:10]):  # Limit
            formatted.append(f"Context {i+1}:\n{ctx[:500]}...")
        return "\n\n".join(formatted)
    
    @trace_function(name="evaluate_rag")
    def evaluate(
        self,
        query: str,
        retrieved_docs: List[Dict[str, Any]],
        answer: str,
        relevant_docs: Optional[List[str]] = None
    ) -> RAGMetrics:
        """
        Run complete RAG evaluation.
        
        Args:
            query: Original query
            retrieved_docs: Retrieved documents
            answer: Generated answer
            relevant_docs: Ground truth relevant docs (optional)
            
        Returns:
            Complete RAGMetrics
        """
        contexts = [doc.get("content", "") for doc in retrieved_docs]
        
        metrics = RAGMetrics()
        
        # Retrieval metrics (if ground truth available)
        if relevant_docs:
            metrics.retrieval = self.evaluate_retrieval(retrieved_docs, relevant_docs)
        
        # Context relevance
        metrics.context_relevance = self.evaluate_context_relevance(query, contexts)
        
        # Context coverage (simple metric)
        if contexts:
            avg_len = sum(len(c) for c in contexts) / len(contexts)
            metrics.context_coverage = min(1.0, avg_len / 500)  # Normalize
        
        # Answer faithfulness
        metrics.answer_faithfulness = self.evaluate_faithfulness(answer, contexts)
        
        # Answer relevance
        metrics.answer_relevance = self.evaluate_answer_relevance(query, answer)
        
        return metrics


def evaluate_with_ragas(
    questions: List[str],
    contexts: List[List[str]],
    answers: List[str],
    ground_truths: Optional[List[str]] = None
) -> Dict[str, float]:
    """
    Evaluate using RAGAS framework (if available).
    
    Args:
        questions: List of questions
        contexts: List of context lists
        answers: Generated answers
        ground_truths: Ground truth answers (optional)
        
    Returns:
        RAGAS metrics dictionary
    """
    try:
        from ragas import evaluate
        from ragas.metrics import (
            faithfulness,
            answer_relevancy,
            context_relevancy,
            context_recall,
        )
        from datasets import Dataset
        
        # Prepare dataset
        data = {
            "question": questions,
            "contexts": contexts,
            "answer": answers,
        }
        
        if ground_truths:
            data["ground_truth"] = ground_truths
        
        dataset = Dataset.from_dict(data)
        
        # Select metrics
        metrics = [faithfulness, answer_relevancy, context_relevancy]
        if ground_truths:
            metrics.append(context_recall)
        
        # Evaluate
        results = evaluate(dataset, metrics=metrics)
        
        return dict(results)
        
    except ImportError:
        logger.warning("RAGAS not installed. Install with: pip install ragas")
        return {}
    except Exception as e:
        logger.error(f"RAGAS evaluation failed: {e}")
        return {}



"""
ReadAnyBook: A RAG-based cheat sheet generator for books and papers.
"""

from setuptools import setup, find_packages
from pathlib import Path

# Read long description from README
this_directory = Path(__file__).parent
long_description = ""
readme_path = this_directory / "README.md"
if readme_path.exists():
    long_description = readme_path.read_text(encoding="utf-8")

# Core dependencies
install_requires = [
    # Configuration
    "pydantic>=2.0.0",
    "pydantic-settings>=2.0.0",
    "PyYAML>=6.0",
    
    # Document processing
    "pypdf>=3.0.0",
    "ebooklib>=0.18",
    "beautifulsoup4>=4.12.0",
    "lxml>=4.9.0",
    "chardet>=5.0.0",
    
    # Embeddings and ML
    "sentence-transformers>=2.2.0",
    "transformers>=4.30.0",
    "torch>=2.0.0",
    
    # Vector stores
    "chromadb>=0.4.0",
    
    # Text processing
    "tiktoken>=0.5.0",
    "rank-bm25>=0.2.2",
    
    # Templating
    "jinja2>=3.1.0",
    
    # Utilities
    "tqdm>=4.65.0",
    "numpy>=1.24.0",
]

# Optional dependencies
extras_require = {
    # CLI and API
    "cli": [
        "typer>=0.9.0",
        "rich>=13.0.0",
    ],
    "api": [
        "fastapi>=0.100.0",
        "uvicorn>=0.23.0",
        "python-multipart>=0.0.6",
    ],
    
    # Additional vector stores
    "qdrant": [
        "qdrant-client>=1.6.0",
    ],
    "weaviate": [
        "weaviate-client>=3.24.0",
    ],
    
    # Additional LLM backends
    "ollama": [
        "ollama>=0.1.0",
    ],
    "openai": [
        "openai>=1.0.0",
    ],
    "vllm": [
        "vllm>=0.2.0",
    ],
    
    # Quantization
    "quantization": [
        "bitsandbytes>=0.41.0",
        "accelerate>=0.24.0",
    ],
    
    # Evaluation
    "eval": [
        "nltk>=3.8.0",
        "rouge-score>=0.1.2",
        "ragas>=0.1.0",
    ],
    
    # Observability
    "observability": [
        "opentelemetry-api>=1.20.0",
        "opentelemetry-sdk>=1.20.0",
        "opentelemetry-exporter-otlp>=1.20.0",
    ],
    
    # Reranking
    "rerank": [
        "sentence-transformers>=2.2.0",  # Already included, but explicit
    ],
    
    # Development
    "dev": [
        "pytest>=7.0.0",
        "pytest-cov>=4.0.0",
        "pytest-asyncio>=0.21.0",
        "black>=23.0.0",
        "isort>=5.12.0",
        "mypy>=1.5.0",
        "ruff>=0.1.0",
    ],
    
    # All optional dependencies
    "all": [
        "typer>=0.9.0",
        "rich>=13.0.0",
        "fastapi>=0.100.0",
        "uvicorn>=0.23.0",
        "python-multipart>=0.0.6",
        "qdrant-client>=1.6.0",
        "weaviate-client>=3.24.0",
        "ollama>=0.1.0",
        "openai>=1.0.0",
        "bitsandbytes>=0.41.0",
        "accelerate>=0.24.0",
        "nltk>=3.8.0",
        "rouge-score>=0.1.2",
        "opentelemetry-api>=1.20.0",
        "opentelemetry-sdk>=1.20.0",
        "opentelemetry-exporter-otlp>=1.20.0",
    ],
}

setup(
    name="readanybook",
    version="0.1.13",
    author="ReadAnyBook Team",
    author_email="team@readanybook.dev",
    description="A RAG-based cheat sheet generator for books and papers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/readanybook/readanybook",
    project_urls={
        "Bug Tracker": "https://github.com/readanybook/readanybook/issues",
        "Documentation": "https://readanybook.dev/docs",
    },
    license="MIT",
    
    packages=find_packages(exclude=["tests", "tests.*", "examples", "examples.*"]),
    package_data={
        "readanybook": [
            "config/*.yaml",
            "templates/*.tex",
        ],
    },
    include_package_data=True,
    
    python_requires=">=3.10",
    install_requires=install_requires,
    extras_require=extras_require,
    
    entry_points={
        "console_scripts": [
            "read-any-book=readanybook.cli.main:main",
            "rab=readanybook.cli.main:main",  # Short alias
        ],
    },
    
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Text Processing :: General",
        "Topic :: Education",
    ],
    
    keywords=[
        "rag",
        "retrieval-augmented-generation",
        "cheat-sheet",
        "pdf",
        "summarization",
        "llm",
        "nlp",
        "education",
    ],
)



"""
Structured logging configuration for ReadAnyBook.

Supports:
- JSON formatted logs for production
- Pretty console logs for development
- Request/document ID tracking
- Integration with observability platforms
"""

import json
import logging
import sys
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional
import contextvars

# Context variables for request tracking
request_id_var: contextvars.ContextVar[str] = contextvars.ContextVar("request_id", default="")
document_id_var: contextvars.ContextVar[str] = contextvars.ContextVar("document_id", default="")
chain_name_var: contextvars.ContextVar[str] = contextvars.ContextVar("chain_name", default="")


class JSONFormatter(logging.Formatter):
    """JSON log formatter for structured logging."""
    
    def format(self, record: logging.LogRecord) -> str:
        log_data = {
            "timestamp": datetime.utcnow().isoformat() + "Z",
            "level": record.levelname,
            "logger": record.name,
            "message": record.getMessage(),
            "module": record.module,
            "function": record.funcName,
            "line": record.lineno,
        }
        
        # Add context variables
        if request_id := request_id_var.get():
            log_data["request_id"] = request_id
        if document_id := document_id_var.get():
            log_data["document_id"] = document_id
        if chain_name := chain_name_var.get():
            log_data["chain_name"] = chain_name
        
        # Add extra fields
        if hasattr(record, "extra_fields"):
            log_data.update(record.extra_fields)
        
        # Add exception info
        if record.exc_info:
            log_data["exception"] = self.formatException(record.exc_info)
        
        return json.dumps(log_data)


class PrettyFormatter(logging.Formatter):
    """Pretty console formatter for development."""
    
    COLORS = {
        "DEBUG": "\033[36m",     # Cyan
        "INFO": "\033[32m",      # Green
        "WARNING": "\033[33m",   # Yellow
        "ERROR": "\033[31m",     # Red
        "CRITICAL": "\033[35m",  # Magenta
    }
    RESET = "\033[0m"
    
    def format(self, record: logging.LogRecord) -> str:
        color = self.COLORS.get(record.levelname, "")
        
        # Build context string
        context_parts = []
        if request_id := request_id_var.get():
            context_parts.append(f"req={request_id[:8]}")
        if document_id := document_id_var.get():
            context_parts.append(f"doc={document_id[:8]}")
        if chain_name := chain_name_var.get():
            context_parts.append(f"chain={chain_name}")
        
        context_str = f" [{', '.join(context_parts)}]" if context_parts else ""
        
        timestamp = datetime.now().strftime("%H:%M:%S.%f")[:-3]
        
        return (
            f"{color}{timestamp} | {record.levelname:8s}{self.RESET} | "
            f"{record.name}:{record.lineno}{context_str} - {record.getMessage()}"
        )


class ContextLogger(logging.LoggerAdapter):
    """Logger adapter that adds context to all log messages."""
    
    def process(self, msg: str, kwargs: Dict[str, Any]) -> tuple:
        extra = kwargs.get("extra", {})
        
        # Add any extra fields to the record
        if "extra_fields" not in extra:
            extra["extra_fields"] = {}
        
        if additional := kwargs.pop("additional", None):
            extra["extra_fields"].update(additional)
        
        kwargs["extra"] = extra
        return msg, kwargs
    
    def with_context(
        self,
        request_id: Optional[str] = None,
        document_id: Optional[str] = None,
        chain_name: Optional[str] = None
    ) -> "ContextLogger":
        """Set context variables for subsequent log calls."""
        if request_id:
            request_id_var.set(request_id)
        if document_id:
            document_id_var.set(document_id)
        if chain_name:
            chain_name_var.set(chain_name)
        return self


_loggers: Dict[str, ContextLogger] = {}
_configured = False


def setup_logging(
    level: str = "INFO",
    log_format: str = "json",
    log_file: Optional[str] = None,
    structured: bool = True
) -> None:
    """
    Set up logging configuration.
    
    Args:
        level: Log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
        log_format: Format type ("json" or "pretty")
        log_file: Optional file path for log output
        structured: Whether to use structured logging
    """
    global _configured
    
    root_logger = logging.getLogger("readanybook")
    root_logger.setLevel(getattr(logging, level.upper()))
    
    # Remove existing handlers
    root_logger.handlers.clear()
    
    # Console handler
    console_handler = logging.StreamHandler(sys.stdout)
    if log_format == "json" and structured:
        console_handler.setFormatter(JSONFormatter())
    else:
        console_handler.setFormatter(PrettyFormatter())
    root_logger.addHandler(console_handler)
    
    # File handler
    if log_file:
        log_path = Path(log_file)
        log_path.parent.mkdir(parents=True, exist_ok=True)
        
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(JSONFormatter())  # Always JSON for files
        root_logger.addHandler(file_handler)
    
    _configured = True


def get_logger(name: str) -> ContextLogger:
    """
    Get a logger with context support.
    
    Args:
        name: Logger name (typically __name__)
        
    Returns:
        ContextLogger instance
    """
    global _configured
    
    if not _configured:
        setup_logging(level="INFO", log_format="pretty", structured=False)
    
    if name not in _loggers:
        logger = logging.getLogger(f"readanybook.{name}" if not name.startswith("readanybook") else name)
        _loggers[name] = ContextLogger(logger, {})
    
    return _loggers[name]


class LogContext:
    """Context manager for scoped logging context."""
    
    def __init__(
        self,
        request_id: Optional[str] = None,
        document_id: Optional[str] = None,
        chain_name: Optional[str] = None
    ):
        self.request_id = request_id
        self.document_id = document_id
        self.chain_name = chain_name
        self._tokens = []
    
    def __enter__(self) -> "LogContext":
        if self.request_id:
            self._tokens.append(("request_id", request_id_var.set(self.request_id)))
        if self.document_id:
            self._tokens.append(("document_id", document_id_var.set(self.document_id)))
        if self.chain_name:
            self._tokens.append(("chain_name", chain_name_var.set(self.chain_name)))
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        for var_name, token in self._tokens:
            if var_name == "request_id":
                request_id_var.reset(token)
            elif var_name == "document_id":
                document_id_var.reset(token)
            elif var_name == "chain_name":
                chain_name_var.reset(token)



"""
Document chunking service for splitting documents into retrievable chunks.

Supports:
- Hierarchical chunking (chapter → section → paragraph)
- Separate streams for narrative, math, and code
- Metadata preservation for grounding
"""

import re
import uuid
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple

from .ingestion import ParsedDocument, ContentBlock
from ..infra.logging import get_logger
from ..infra.tracing import trace_function

logger = get_logger(__name__)


class ChunkType(Enum):
    """Types of content chunks."""
    NARRATIVE = "narrative"  # Regular text, explanations, definitions
    MATH = "math"            # Mathematical formulas and equations
    CODE = "code"            # Code blocks, algorithms, pseudocode
    HEADING = "heading"      # Section headings
    MIXED = "mixed"          # Contains multiple types


@dataclass
class Chunk:
    """Represents a document chunk for indexing and retrieval."""
    id: str
    content: str
    chunk_type: ChunkType
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    # Source information
    document_id: Optional[str] = None
    page_numbers: List[int] = field(default_factory=list)
    section_hierarchy: List[str] = field(default_factory=list)
    
    # Position information
    start_char: Optional[int] = None
    end_char: Optional[int] = None
    chunk_index: int = 0
    
    @classmethod
    def create(
        cls,
        content: str,
        chunk_type: ChunkType,
        **kwargs
    ) -> "Chunk":
        """Create a new chunk with auto-generated ID."""
        return cls(
            id=str(uuid.uuid4()),
            content=content,
            chunk_type=chunk_type,
            **kwargs
        )
    
    @property
    def token_estimate(self) -> int:
        """Rough estimate of token count (words * 1.3)."""
        return int(len(self.content.split()) * 1.3)
    
    @property
    def char_count(self) -> int:
        """Character count of content."""
        return len(self.content)
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert chunk to dictionary for storage."""
        return {
            "id": self.id,
            "content": self.content,
            "chunk_type": self.chunk_type.value,
            "document_id": self.document_id,
            "page_numbers": self.page_numbers,
            "section_hierarchy": self.section_hierarchy,
            "chunk_index": self.chunk_index,
            **self.metadata
        }


class TextSplitter:
    """Base text splitter with configurable separators."""
    
    def __init__(
        self,
        chunk_size: int = 1000,
        chunk_overlap: int = 200,
        separators: Optional[List[str]] = None,
        length_function: Optional[callable] = None
    ):
        """
        Initialize the text splitter.
        
        Args:
            chunk_size: Target size for chunks (in characters)
            chunk_overlap: Overlap between consecutive chunks
            separators: List of separators to try, in order of preference
            length_function: Function to measure text length
        """
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.separators = separators or ["\n\n", "\n", ". ", " "]
        self.length_function = length_function or len
    
    def split_text(self, text: str) -> List[str]:
        """
        Split text into chunks.
        
        Args:
            text: Text to split
            
        Returns:
            List of text chunks
        """
        return self._split_text_recursive(text, self.separators)
    
    def _split_text_recursive(self, text: str, separators: List[str]) -> List[str]:
        """Recursively split text using available separators."""
        if not separators:
            # No separators left, split by chunk size
            return self._split_by_size(text)
        
        separator = separators[0]
        remaining_separators = separators[1:]
        
        # Split by current separator
        splits = text.split(separator)
        
        chunks = []
        current_chunk = []
        current_length = 0
        
        for split in splits:
            split_length = self.length_function(split)
            
            # If single split is too large, recursively split it
            if split_length > self.chunk_size:
                # First, add current accumulated chunk if any
                if current_chunk:
                    chunks.append(separator.join(current_chunk))
                    current_chunk = []
                    current_length = 0
                
                # Recursively split the large segment
                sub_chunks = self._split_text_recursive(split, remaining_separators)
                chunks.extend(sub_chunks)
                continue
            
            # Check if adding this split exceeds chunk size
            new_length = current_length + split_length + (len(separator) if current_chunk else 0)
            
            if new_length > self.chunk_size and current_chunk:
                # Save current chunk and start new one
                chunks.append(separator.join(current_chunk))
                
                # Apply overlap
                overlap_splits = self._get_overlap_splits(current_chunk, separator)
                current_chunk = overlap_splits + [split]
                current_length = self.length_function(separator.join(current_chunk))
            else:
                current_chunk.append(split)
                current_length = new_length
        
        # Add final chunk
        if current_chunk:
            chunks.append(separator.join(current_chunk))
        
        return [c for c in chunks if c.strip()]
    
    def _split_by_size(self, text: str) -> List[str]:
        """Split text by size when no separators work."""
        chunks = []
        start = 0
        
        while start < len(text):
            end = start + self.chunk_size
            
            # Try to break at a word boundary
            if end < len(text):
                space_idx = text.rfind(' ', start, end)
                if space_idx > start:
                    end = space_idx
            
            chunks.append(text[start:end].strip())
            start = end - self.chunk_overlap
        
        return chunks
    
    def _get_overlap_splits(self, splits: List[str], separator: str) -> List[str]:
        """Get splits for overlap from the end of current chunk."""
        if not splits or self.chunk_overlap <= 0:
            return []
        
        overlap_splits = []
        current_length = 0
        
        for split in reversed(splits):
            split_length = self.length_function(split) + len(separator)
            if current_length + split_length > self.chunk_overlap:
                break
            overlap_splits.insert(0, split)
            current_length += split_length
        
        return overlap_splits


class MathAwareTextSplitter(TextSplitter):
    """Text splitter that preserves mathematical expressions."""
    
    # Patterns for math content
    MATH_PATTERNS = [
        (r'\$\$[\s\S]*?\$\$', 'display'),           # Display math
        (r'\$[^\$]+\$', 'inline'),                   # Inline math
        (r'\\begin\{equation\}[\s\S]*?\\end\{equation\}', 'equation'),
        (r'\\begin\{align\*?\}[\s\S]*?\\end\{align\*?\}', 'align'),
        (r'\\begin\{eqnarray\*?\}[\s\S]*?\\end\{eqnarray\*?\}', 'eqnarray'),
        (r'\\begin\{gather\*?\}[\s\S]*?\\end\{gather\*?\}', 'gather'),
        (r'\\\[[\s\S]*?\\\]', 'display_bracket'),     # \[ \] math
        (r'\\\([\s\S]*?\\\)', 'inline_bracket'),     # \( \) math
    ]
    
    def __init__(self, *args, preserve_math: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self.preserve_math = preserve_math
    
    def split_text(self, text: str) -> List[str]:
        """Split text while preserving math blocks."""
        if not self.preserve_math:
            return super().split_text(text)
        
        # Replace math blocks with placeholders
        math_blocks = {}
        protected_text = text
        
        for pattern, math_type in self.MATH_PATTERNS:
            for i, match in enumerate(re.finditer(pattern, protected_text)):
                placeholder = f"__MATH_{len(math_blocks)}__"
                math_blocks[placeholder] = match.group(0)
                protected_text = protected_text.replace(match.group(0), placeholder, 1)
        
        # Split the protected text
        chunks = super().split_text(protected_text)
        
        # Restore math blocks
        restored_chunks = []
        for chunk in chunks:
            for placeholder, math_content in math_blocks.items():
                chunk = chunk.replace(placeholder, math_content)
            restored_chunks.append(chunk)
        
        return restored_chunks


class CodeAwareTextSplitter(TextSplitter):
    """Text splitter that preserves code blocks."""
    
    CODE_PATTERNS = [
        (r'```[\s\S]*?```', 'fenced'),              # Fenced code blocks
        (r'\\begin\{lstlisting\}[\s\S]*?\\end\{lstlisting\}', 'lstlisting'),
        (r'\\begin\{verbatim\}[\s\S]*?\\end\{verbatim\}', 'verbatim'),
        (r'\\begin\{algorithm\}[\s\S]*?\\end\{algorithm\}', 'algorithm'),
        (r'\\begin\{minted\}[\s\S]*?\\end\{minted\}', 'minted'),
    ]
    
    def __init__(self, *args, preserve_code: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self.preserve_code = preserve_code
    
    def split_text(self, text: str) -> List[str]:
        """Split text while preserving code blocks."""
        if not self.preserve_code:
            return super().split_text(text)
        
        # Replace code blocks with placeholders
        code_blocks = {}
        protected_text = text
        
        for pattern, code_type in self.CODE_PATTERNS:
            for match in re.finditer(pattern, protected_text):
                placeholder = f"__CODE_{len(code_blocks)}__"
                code_blocks[placeholder] = match.group(0)
                protected_text = protected_text.replace(match.group(0), placeholder, 1)
        
        # Split the protected text
        chunks = super().split_text(protected_text)
        
        # Restore code blocks
        restored_chunks = []
        for chunk in chunks:
            for placeholder, code_content in code_blocks.items():
                chunk = chunk.replace(placeholder, code_content)
            restored_chunks.append(chunk)
        
        return restored_chunks


class HierarchicalChunker:
    """
    Hierarchical chunking that respects document structure.
    
    Creates chunks at multiple levels:
    - Document level summaries
    - Chapter/section chunks
    - Paragraph-level chunks
    """
    
    def __init__(
        self,
        chunk_size: int = 1000,
        chunk_overlap: int = 200,
        min_chunk_size: int = 100,
        max_chunk_size: int = 2000,
        preserve_math: bool = True,
        preserve_code: bool = True
    ):
        """
        Initialize the hierarchical chunker.
        
        Args:
            chunk_size: Target chunk size
            chunk_overlap: Overlap between chunks
            min_chunk_size: Minimum chunk size
            max_chunk_size: Maximum chunk size
            preserve_math: Preserve math expressions
            preserve_code: Preserve code blocks
        """
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.min_chunk_size = min_chunk_size
        self.max_chunk_size = max_chunk_size
        
        # Create specialized splitter
        self.splitter = MathAwareTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            preserve_math=preserve_math
        )
        
        self.preserve_math = preserve_math
        self.preserve_code = preserve_code
    
    def chunk_document(self, document: ParsedDocument) -> List[Chunk]:
        """
        Chunk a parsed document hierarchically.
        
        Args:
            document: ParsedDocument to chunk
            
        Returns:
            List of Chunks with metadata
        """
        chunks = []
        
        # Group content blocks by section
        sections = self._group_by_section(document.content_blocks)
        
        # Process each section
        for section_path, blocks in sections.items():
            section_chunks = self._chunk_section(
                blocks,
                section_hierarchy=list(section_path) if section_path else [],
                document_id=document.metadata.source_path
            )
            chunks.extend(section_chunks)
        
        # Assign chunk indices
        for i, chunk in enumerate(chunks):
            chunk.chunk_index = i
        
        logger.info(f"Created {len(chunks)} chunks from document")
        return chunks
    
    def _group_by_section(
        self, 
        blocks: List[ContentBlock]
    ) -> Dict[Tuple[str, ...], List[ContentBlock]]:
        """Group content blocks by their section hierarchy."""
        sections: Dict[Tuple[str, ...], List[ContentBlock]] = {}
        
        for block in blocks:
            section_key = tuple(block.section_hierarchy) if block.section_hierarchy else ()
            if section_key not in sections:
                sections[section_key] = []
            sections[section_key].append(block)
        
        return sections
    
    def _chunk_section(
        self,
        blocks: List[ContentBlock],
        section_hierarchy: List[str],
        document_id: Optional[str]
    ) -> List[Chunk]:
        """Chunk a single section's content."""
        chunks = []
        
        # Separate blocks by type
        narrative_blocks = []
        math_blocks = []
        code_blocks = []
        
        for block in blocks:
            if block.is_math:
                math_blocks.append(block)
            elif block.is_code:
                code_blocks.append(block)
            else:
                narrative_blocks.append(block)
        
        # Chunk narrative content
        if narrative_blocks:
            narrative_text = "\n\n".join(b.content for b in narrative_blocks)
            page_numbers = list(set(
                b.page_number for b in narrative_blocks if b.page_number
            ))
            
            text_chunks = self.splitter.split_text(narrative_text)
            
            for text in text_chunks:
                if len(text) >= self.min_chunk_size:
                    chunks.append(Chunk.create(
                        content=text,
                        chunk_type=ChunkType.NARRATIVE,
                        document_id=document_id,
                        section_hierarchy=section_hierarchy,
                        page_numbers=page_numbers
                    ))
        
        # Create separate chunks for math content
        for block in math_blocks:
            chunks.append(Chunk.create(
                content=block.content,
                chunk_type=ChunkType.MATH,
                document_id=document_id,
                section_hierarchy=section_hierarchy,
                page_numbers=[block.page_number] if block.page_number else [],
                metadata=block.metadata
            ))
        
        # Create separate chunks for code content
        for block in code_blocks:
            chunks.append(Chunk.create(
                content=block.content,
                chunk_type=ChunkType.CODE,
                document_id=document_id,
                section_hierarchy=section_hierarchy,
                page_numbers=[block.page_number] if block.page_number else [],
                metadata=block.metadata
            ))
        
        return chunks


class ChunkingService:
    """
    Main chunking service for document processing.
    
    Provides different chunking strategies based on document type.
    """
    
    def __init__(self, settings=None):
        """
        Initialize the chunking service.
        
        Args:
            settings: Optional ChunkingSettings
        """
        self.settings = settings
        
        # Default settings
        chunk_size = settings.chunk_size if settings else 1000
        chunk_overlap = settings.chunk_overlap if settings else 200
        min_chunk_size = settings.min_chunk_size if settings else 100
        max_chunk_size = settings.max_chunk_size if settings else 2000
        preserve_math = settings.preserve_math if settings else True
        preserve_code = settings.preserve_code if settings else True
        
        self.hierarchical_chunker = HierarchicalChunker(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            min_chunk_size=min_chunk_size,
            max_chunk_size=max_chunk_size,
            preserve_math=preserve_math,
            preserve_code=preserve_code
        )
        
        self.text_splitter = MathAwareTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            preserve_math=preserve_math
        )
    
    @trace_function(name="chunk_document")
    def chunk(
        self,
        document: ParsedDocument,
        strategy: str = "hierarchical"
    ) -> List[Chunk]:
        """
        Chunk a document using specified strategy.
        
        Args:
            document: ParsedDocument to chunk
            strategy: Chunking strategy ("hierarchical", "simple", "by_type")
            
        Returns:
            List of Chunks
        """
        if strategy == "hierarchical":
            return self.hierarchical_chunker.chunk_document(document)
        elif strategy == "simple":
            return self._simple_chunk(document)
        elif strategy == "by_type":
            return self._chunk_by_type(document)
        else:
            raise ValueError(f"Unknown chunking strategy: {strategy}")
    
    def _simple_chunk(self, document: ParsedDocument) -> List[Chunk]:
        """Simple flat chunking of document text."""
        chunks = []
        text_chunks = self.text_splitter.split_text(document.raw_text)
        
        for i, text in enumerate(text_chunks):
            chunks.append(Chunk.create(
                content=text,
                chunk_type=ChunkType.MIXED,
                document_id=document.metadata.source_path,
                chunk_index=i
            ))
        
        return chunks
    
    def _chunk_by_type(self, document: ParsedDocument) -> List[Chunk]:
        """Chunk document with separate streams by content type."""
        chunks = []
        
        # Collect content by type
        narrative_content = []
        math_content = []
        code_content = []
        
        for block in document.content_blocks:
            if block.is_math:
                math_content.append(block)
            elif block.is_code:
                code_content.append(block)
            elif not block.is_heading:
                narrative_content.append(block)
        
        # Chunk narrative content
        narrative_text = "\n\n".join(b.content for b in narrative_content)
        for text in self.text_splitter.split_text(narrative_text):
            chunks.append(Chunk.create(
                content=text,
                chunk_type=ChunkType.NARRATIVE,
                document_id=document.metadata.source_path
            ))
        
        # Keep math blocks as individual chunks
        for block in math_content:
            chunks.append(Chunk.create(
                content=block.content,
                chunk_type=ChunkType.MATH,
                document_id=document.metadata.source_path,
                section_hierarchy=block.section_hierarchy
            ))
        
        # Keep code blocks as individual chunks
        for block in code_content:
            chunks.append(Chunk.create(
                content=block.content,
                chunk_type=ChunkType.CODE,
                document_id=document.metadata.source_path,
                section_hierarchy=block.section_hierarchy
            ))
        
        # Assign indices
        for i, chunk in enumerate(chunks):
            chunk.chunk_index = i
        
        return chunks
    
    def chunk_text(self, text: str, chunk_type: ChunkType = ChunkType.MIXED) -> List[Chunk]:
        """
        Chunk raw text.
        
        Args:
            text: Text to chunk
            chunk_type: Type to assign to chunks
            
        Returns:
            List of Chunks
        """
        chunks = []
        text_chunks = self.text_splitter.split_text(text)
        
        for i, content in enumerate(text_chunks):
            chunks.append(Chunk.create(
                content=content,
                chunk_type=chunk_type,
                chunk_index=i
            ))
        
        return chunks
